"""
File Purpose: EbysusCalculator
"""
import os

from .ebysus_bases import EbysusBasesLoader
from .ebysus_collisions import EbysusCollisionsLoader
from .ebysus_dimensions import EbysusFluidList
from .ebysus_efield import EbysusEfieldLoader
from .ebysus_temperatures import EbysusTemperaturesLoader
from ...dimensions import SnapList
from ...errors import LoadingNotImplementedError
from ...plasma_calculator import MultifluidPlasmaCalculator
from ...tools import alias, alias_child, simple_property
from ...defaults import DEFAULTS


class EbysusCalculator(EbysusBasesLoader, EbysusCollisionsLoader,
                       EbysusEfieldLoader, EbysusTemperaturesLoader,
                       MultifluidPlasmaCalculator):
    '''PlasmaCalculator for Ebysus.

    The idea is to use EbysusData (from helita.sim.ebysus) to read files.
        But hook everything up so that PlasmaCalcs does all the heavy-lifting.

    dd should be an EbysusData object.

    collisions_mode: str, default 'best'
        default mode for getting collision frequencies;
        'helita' --> get collision frequencies directly from helita.
        for more options/details, see help(type(self).collisions_mode).
    match_type: 'AUX', 'PHYSICS', 1, 0, or None
        1 or 'AUX' --> set dd.match_type = MATCH_AUX
        0 or 'PHYSICS' --> set dd.match_type = MATCH_PHYSICS
        None --> don't adjust dd.match_type.

    UNITS note: right now, units is fully managed by dd;
        self.u(conversion) will always be 1 (e.g. self.u('l') == self.u('m') == ... == 1);
        self.units == 'si' --> dd.units_output == 'si';
        self.units == 'raw' --> dd.units_output == 'simu'.
        [TODO] handle units via PlasmaCalcs instead of dd.
    '''
    quasineutral = True   # Ebysus is always quasineutral.

    def __init__(self, dd, *, collisions_mode='best', match_type='AUX', **kw_super):
        self.dd = dd
        self.collisions_mode = collisions_mode
        if match_type is not None:
            dd.match_type = {1: 1, 'AUX': 1, 0: 0, 'PHYSICS': 0}[match_type]
        self._setup_on_set_v_methods()
        snapsdict = self.get_init_snaps()
        fdict = self.get_init_fluids()
        kw = {**snapsdict, **fdict, **kw_super}
        kw['units'] = kw.get('units', dd.units_output)
        super().__init__(**kw)

    dirname = alias_child('dd', 'fdir')

    title = simple_property('_title', setdefaultvia='_default_title',
            doc='''title to help distinguish this calculator from others.
            E.g. might want to add self.title to plots. Default: os.path.basename(self.dirname).''')
    def _default_title(self):
        '''default title for self. Here, just returns os.path.basename(self.dirname)'''
        return os.path.basename(self.dirname)

    # # # UNITS SETUP # # #
    @property
    def units(self):
        '''units, but also set dd.units_output when setting.
        UNITS note: right now, units is fully managed by dd;
            self.u(conversion) will always be 1 (e.g. self.u('l') == self.u('m') == ... == 1);
            self.units == 'si' --> dd.units_output == 'si';
            self.units == 'raw' --> dd.units_output == 'simu'.
        '''
        return super().units
    @units.setter
    def units(self, value):
        if value == 'simu':
            value = 'raw'
        super(EbysusCalculator, type(self)).units.fset(self, value)
        self.dd.units_output = 'simu' if (value == 'raw') else value

    # # # DIMENSIONS SETUP -- MAINDIMS # # #
    maindims = ('x', 'y', 'z')

    def get_maindims_coords(self):
        '''return dict of {'x': xcoords, 'y': ycoords, 'z': zcoords}, in dd.units_output units.
        coords will be sliced according to self.slices.
        '''
        result = dict(x=self.dd.get_coord('x'),
                        y=self.dd.get_coord('y'),
                        z=self.dd.get_coord('z'))
        if self.flip_z_mesh:
            result['z'] = -result['z']
        return self._apply_maindims_slices_to_dict(result)

    cls_behavior_attrs.register('flip_z_mesh', default=False)
    flip_z_mesh = simple_property('_flip_z_mesh', default=False,
            doc='''whether to flip (multiply by -1) the z mesh coordinates relative to meshfile.
            When True, z<0 implies "below photosphere", for solar simulations.''')

    get_space_coords = alias('get_maindims_coords')

    def maindims_with_extent(self):
        '''return tuple of maindims with length > 1.
        ebysus output is all technically 3D but some dims might have length 1 for 2D simulation.
        It can be convenient to know which dims have extent / which do not.
        '''
        return tuple(x for x in self.maindims if getattr(self.dd, f'n{x}') > 1)

    def maindims_with_no_extent(self):
        '''return tuple of maindims with length == 1.
        ebysus output is all technically 3D but some dims might have length 1 for 2D simulation.
        It can be convenient to know which dims have extent / which do not.
        '''
        return tuple(x for x in self.maindims if getattr(self.dd, f'n{x}') == 1)

    # # # DIMENSIONS SETUP -- OTHERS # # #
    def match_dd_dims(self):
        '''matches self.snap, fluid, and jfluid to self.dd.snap, ifluid, and jfluid.'''
        dd = self.dd
        try:
            snaps = dd.snaps
        except TypeError:  # dd.snap is a single snapshot
            self.snap = str(dd.snap)
        else:  # dd.snap is multiple snapshots
            self.snap = [str(snap) for snap in snaps]
        self.fluid = dd.get_fluid_name(dd.ifluid)
        self.jfluid = dd.get_fluid_name(dd.jfluid)

    def get_init_snaps(self):
        '''return dict of snaps, snap for setting self.snaps & self.snap.'''
        dd = self.dd
        with dd.maintaining('snap'):  # restore dd.snap afterwards
            dd[:]  # sets dd.snap to dd.get_snaps()
            snaps = [str(snap) for snap in dd.snaps]
            times_raw = dd.get_coord('t', 'simu')
        snaplist = SnapList.from_lists(s=snaps, t=times_raw)
        return dict(snaps=snaplist, snap=0)

    def get_init_fluids(self):
        '''return dict of fluids, fluid, jfluids, jfluid for setting those attrs of self.'''
        dd = self.dd
        fluids = dd.fluids
        names = fluids.name
        masses = fluids.atomic_weight  # [amu]
        charges = fluids.ionization   # [elementary charge units]
        SLs = fluids.SL
        electrons = dict(name='e', m= dd.get_mass((-1,0), units='amu'), q= -1, SL=(-1,0))
        others = [dict(name=name_, m=m_, q=q_, SL=SL_) for name_, m_, q_, SL_ in zip(names, masses, charges, SLs)]
        fluidlist = EbysusFluidList.from_dicts([electrons] + others)
        return dict(fluids=fluidlist, fluid=dd.get_fluid_name(dd.ifluid),
                    jfluids=fluidlist, jfluid=dd.get_fluid_name(dd.jfluid))

    def _setup_on_set_v_methods(self):
        '''sets up methods so that when dims in self are set, appropriate dims in self.dd are set too.
        E.g., when setting self.fluid, set self.dd.fluid too, if self.fluid_is_iterable().
        '''
        self._setup_on_set_snap()
        self._setup_on_set_fluid()
        self._setup_on_set_jfluid()

    def _setup_on_set_snap(self):
        '''sets self.snap_dim.on_set_v = method which also sets dd.snap.'''
        def on_set_snap(value):
            '''when setting snap, also set dd.snap.'''
            try:
                snap_s = self.snap.file_s(self)
                # ^^ file_s == "the str used to get file for this snap & calculator"
                # ^^ or, if snap is a SnapList, it should be an array of file_s for each snap in the list.
            except AttributeError:
                if DEFAULTS.DEBUG > 3:
                    raise
                return   # don't know what to set snap to...
            if isinstance(snap_s, str):
                self.dd.snap = int(snap_s)
            else:  # snap_s is a list of file_s values
                self.dd.snap = [int(s) for s in snap_s]
        self.snap_dim.on_set_v = on_set_snap

    def _setup_on_set_fluid(self):
        '''sets self.fluid_dim.on_set_v = method which sets dd.ifluid if setting fluid to single value.'''
        def on_set_fluid(value):
            '''when setting fluid, also set dd.ifluid, if not fluid_is_iterable().'''
            if not self.fluid_is_iterable():
                self.dd.ifluid = self.fluid.SL
        self.fluid_dim.on_set_v = on_set_fluid

    def _setup_on_set_jfluid(self):
        '''sets self.jfluid_dim.on_set_v = method which sets dd.jfluid if setting jfluid to single value.'''
        def on_set_jfluid(value):
            '''when setting jfluid, also set dd.jfluid, if not jfluid_is_iterable().'''
            if not self.jfluid_is_iterable():
                self.dd.jfluid = self.jfluid.SL
        self.jfluid_dim.on_set_v = on_set_jfluid

    # # # LOADING VARS # # #
    def _var_dimmed(self, ebysus_var):
        '''return ebysus_var, possibly slightly adjusted based on currently-loading dimension(s).
        Here:
            append component AND 'c' if self._loading_component
        E.g. 'b' turns into 'bzc' when loading 'z' component.
        '''
        result = ebysus_var
        if getattr(self, '_loading_component', False):
            result = result + str(self.component) + 'c'
        return result

    def load_direct(self, var, *args, **kw):
        '''load var "directly", either from a file or from self.direct_overrides.
        first, add dimensions to var if appropriate, based on the currently-loading dimension(s).
            E.g. 'b' turns into 'bzc' when loading 'z' component.
        If there is is an override for that var, use it; otherwise load from file.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        ebysus_var = self._var_dimmed(var)
        return super().load_direct(ebysus_var, *args, **kw)

    def load_fromfile(self, ebysus_var, *args, **kw):
        '''load ebysus_var directly from self.dd.

        ebysus_var: str
            the name of the variable to read. Should include all dimensions as appropriate.
            E.g., use 'bzc', not 'b', to get magnetic field z component.

        End-user should not need to worry about this function;
            it's mainly here for connecting the ebysus base vars for PlasmaCalcs appropriately.
        '''
        try:
            return self.dd(ebysus_var, *args, **kw)
        except ValueError as err:
            raise LoadingNotImplementedError(f'{type(self).__name__}.load_fromfile({ebysus_var!r})') from err

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''return list of strings to use inside self.__repr__'''
        contents = super()._repr_contents()
        contents = [f'dirname={self.dirname!r}'] + list(contents)
        return contents



"""
File Purpose: Dimension
"""

import builtins

from .dimension_value import DimensionSpecialValueSpecifier

from ...errors import DimensionError, DimensionKeyError, DimensionValueError
from ...tools import (
    simple_property,
    is_iterable_dim, take_along_dimension, join_along_dimension, xarray_assign,
    format_docstring,
    alias_child,
    using_attrs, maintaining_attrs,
    UNSET,
)


''' --------------------- Dimension --------------------- '''

_paramdocs_dimension = {
    'v': '''v: None, DimensionValue, or DimensionValueList
                current value of this dimension. Might be multiple values.
                None --> use self.values instead.
                note: when setting self.v = value, will attempt to match value(s) from self.values if possible:
                    if value is None, set self.v = None.
                    else, if self.values exists (and is not None), attempt self.v = self.values.get(v).
                    By default (if the previous cases fail or don't hold), sets self.v = value.''',
    'values': '''values: UNSET, None, or DimensionValueList
                all possible values of this dimension.
                UNSET (or, del self.values) --> raise an AttributeError when trying to access.
                None --> disables the AttributeError, but also disables the "smart" behavior of self.v,
                        where setting self.v tries to match value(s) from self.values.''',
    'on_set_v': '''on_set_v: None or callable of 1 argument.
                if not None, call self.on_set_v(value) immediately after doing self.v = value.''',
    }

@format_docstring(**_paramdocs_dimension)
class Dimension():
    '''a single dimension, representing a current value AND a list of all possible values.

    when creating a subclass, provide:
            name: None or str
                name of this dimension. E.g. 'fluid'.
                None --> this will not be a useable dimension.
                        (use None if creating a subclass which is intended to be subclassed again)
            plural: None or str
                name of this dimension, in plural form. E.g. 'fluids'.
                None --> use str(name)+'s'.
            value_error_type: None or type
                type for cls.dimension_value_error. E.g. FluidValueError
                None --> use DimensionValueError
            key_error_type: None or type
                type for cls.dimension_key_error. E.g. FluidKeyError
                None --> use DimensionKeyError

        Note that docstrings for DIM_METHODS will be formatted using these names,
            via docstring.format(dim=name, dims=plural), if name is provided.


    when creating an instance (possibly an instance of a subclass), can provide:
            {v}
            {values}
            {on_set_v}


    --- Example ---
        class FluidDimension(Dimension, name='fluid'):
            pass   # no special methods for this example

        @FluidDimension.setup_haver  # setup FluidHaver by attaching various relevant properties & methods.
        class FluidHaver(DimensionHaver, dimension='fluid'):
            pass   # no special methods for this example

        after the lines above, FluidHaver will "have" a fluid dimension, e.g.:
            - fluid_haver.fluid_dim       --> FluidDimension instance associated with this fluid_haver instance
            - fluid_haver.fluid           --> fluid_haver.fluid_dim.v
            - fluid_haver.fluids          --> fluid_haver.fluid_dim.values
            - fluid_haver.current_n_fluid --> fluid_haver.fluid_dim.current_n
            - fluid_haver.iter_fluids     --> fluid_haver.fluid_dim.iter_values
            - fluid_haver.take_fluid      --> fluid_haver.fluid_dim.take
        where fluid_haver is an instance of FluidHaver.
        Note that FluidHaver has those attributes as well but they are properties or methods, e.g.:
            FluidHaver.fluid       = property(lambda self: self.fluid_dim.v)             # roughly.
            FluidHaver.iter_fluids = property(lambda self: self.fluid_dim.iter_values)   # roughly.

        For a complete list of all methods which will be attached, see Dimension.setup_haver and DIM_METHODS.
    '''
    name = None
    plural = None
    dimension_value_error = DimensionValueError
    dimension_key_error = DimensionKeyError

    def __init__(self, v=None, values=UNSET, *, on_set_v=None, **kw_super):
        super().__init__(**kw_super)
        self.on_set_v = on_set_v
        if values is not UNSET: self.values = values
        self.v = v  # ^note: set values before v, because v setter might use self.values.

    def __init_subclass__(cls, *, name=None, plural=None,
                          value_error_type=None, key_error_type=None,
                          **kw_super):
        '''setup cls based on provided name and plural.

        name: None or str
            if provided, does the following:
            - sets cls.name
            - sets cls.plural
            - replace DIM_METHODS keys with key.format(dim=name, dims=plural).
        plural: None or str
            plural form of name. If None, use str(name)+'s'. (Only used if name is not None)
        value_error_type: None or type
            if provided, set cls.dimension_value_error = value_error_type
        key_error_type: None or type
            if provided, set cls.dimension_key_error = key_error_type
        '''
        super().__init_subclass__(**kw_super)
        if name is not None:
            cls.name = name
            if plural is None: plural = f'{name}s'
            cls.plural = plural
            dim_methods_orig = cls.DIM_METHODS
            cls.DIM_METHODS = dict()  # new dict, to avoid overwriting the old dict.
            for key, method in dim_methods_orig.items():
                cls.DIM_METHODS[key.format(dim=name, dims=plural)] = method
                # note: difficult to format docstring;
                #   having different docstring requires having multiple copies of the method,
                #   which makes the code less clean. That is why the next line is commented-out.
                # method.__doc__ = method.__doc__.format(dim=name, dims=plural)
        if value_error_type is not None: cls.dimension_value_error = value_error_type
        if key_error_type is not None: cls.dimension_key_error = key_error_type

    # # # SETUP IN ANOTHER CLASS -- TRACKING DIM_METHODS # # #
    @classmethod
    def setup_haver(cls, haver_cls):
        '''setup haver_cls, so that it "has" this dimension,
        in the sense that the following are defined (replacing "fluid" with cls.name):

        - h = haver_cls
        - plural = cls.plural
        - h.fluid_dim_cls  = cls
        - h.fluid_dim      = alias to the instance of cls associated with an instance of h
        - h.fluid          = alias to h.fluid_dim.v
        - h.{plural}       = alias to h.fluid_dim.values
        - h._loading_fluid = alias to h.fluid_dim.loading
        - for (key, method) in cls.DIM_METHODS.items():
            # assert that h.{key} doesn't exist yet, then define:
            h.{key}        = alias to h.fluid_dim.{method}
            # (e.g., h.assign_snap_coord = alias to h.snap_dim.assign_coord)

        Also, call haver_cls.__setup_haver__(cls) if it exists.

        returns haver_cls, so that this function may be used as a class decorator.
        '''
        h = haver_cls
        dim = cls.name   # e.g. 'fluid'
        dim_dim = f'{dim}_dim'  # e.g. 'fluid_dim'
        dim_dim_cls = f'{dim}_dim_cls'   # e.g. 'fluid_dim_cls'

        # e.g. h.fluid_dim_cls = cls
        assert not hasattr(h, dim_dim_cls), f"haver_cls.{dim_dim_cls} already exists!"
        setattr(h, dim_dim_cls, cls)

        # e.g. h.fluid_dim --> the instance of cls which is associated with the instance of h.
        # if h.{dim_dim} hasn't been set, use h.{dim_dim} = h.{dim_dim_cls}().
        assert not hasattr(h, dim_dim), f"haver_cls.{dim_dim} already exists!"
        h_dim_dim = simple_property(f'_{dim_dim}', setdefault=getattr(h, dim_dim_cls),
                                    doc=f'''{dim} dimension for {h.__name__}.''')
        setattr(h, dim_dim, h_dim_dim)

        # e.g. h.fluid --> alias to h.fluid_dim.v
        assert not hasattr(h, dim), f"haver_cls.{dim} already exists!"
        setattr(h, dim, alias_child(dim_dim, 'v'))

        # e.g. h.fluids --> alias to h.fluid_dim.values
        assert not hasattr(h, cls.plural), f"haver_cls.{cls.plural} already exists!"
        setattr(h, cls.plural, alias_child(dim_dim, 'values'))

        # e.g. h._loading_fluid --> alias to h.fluid_dim.loading
        assert not hasattr(h, f'_loading_{dim}'), f"haver_cls._loading_{dim} already exists!"
        setattr(h, f'_loading_{dim}', alias_child(dim_dim, 'loading'))

        # e.g. h.iter_fluid --> alias to h.fluid_dim.iter   # for each method in cls.DIM_METHODS
        for fmtname, methodname in cls.DIM_METHODS.items():
            assert not hasattr(h, fmtname), f"haver_cls.{fmtname} already exists!"
            setattr(h, fmtname, alias_child(dim_dim, methodname))

        # call haver_cls.__setup_haver__(cls) if it exists.
        try:
            h_setup_haver = h.__setup_haver__
        except AttributeError:
            pass
        else:
            h_setup_haver(cls)

        # return haver_cls, so that this function may be used as a class decorator.
        return h


    DIM_METHODS = dict()  # dict of methods to use in cls.setup_haver_cls; {fmtname: method}

    def _dim_method(fmtname, dim_methods=DIM_METHODS):
        '''return decorator(f) which returns f after putting {fmtname: f.__name__} into dim_methods'''
        def decorator(f):
            dim_methods[fmtname] = f.__name__
            return f
        return decorator

    # # # CONVENIENCE # # #
    using = using_attrs
    maintaining = maintaining_attrs

    # # # PROPERTIES # # #
    @property
    def v(self):
        '''current value in self.
        Getting self.v:
            Return self._v if it exists (and is not None), else return self.values.
        Setting self.v = value:
            Attempt to match value from self.values if possible:
                if value is None, set self.v = None
                if value is a DimensionSpecialValueSpecifier, call value.value_to_set(self).
                else, if self.values exists (and is not None), attempt self.v = self.values.get(value).
                By default (if the previous cases fail or don't hold), sets self.v = value.
            Afterwards, call self.on_set_v(value), if self.on_set_v is not None.
        '''
        v = getattr(self, '_v', None)
        return self.values if v is None else v
    @v.setter
    def v(self, val):
        '''set self.v, matching value from self.values if possible.
        See help(type(self).v) for details.
        '''
        if val is None:
            result = val
        elif isinstance(val, DimensionSpecialValueSpecifier):
            result = val.getter(self.values)
        else:
            values = self.values   # <-- will raise AttributeError if self.values doesn't exist!
            if values is None:
                result = val
            else:  # self.values is not None; attempt to use self.values.get(val).
                try:
                    self_values_get = self.values.get
                except AttributeError:
                    result = val  # self.values doesn't have a 'get' method! (it's not a DimensionValueList.)
                else: 
                    result = self_values_get(val)
        self._v = result
        if self.on_set_v is not None:
            self.on_set_v(val)

    @property
    def values(self):
        '''all possible values of this dimension.
        If values does not exist, raise an AttributeError with a helpful message.
            (the message is most helpful at "high level", when using DimensionHaver,
            so it uses the dimension's name instead of 'v' and 'values'.)
        If you want to avoid using DimensionValueList (at the cost of not using its smart behavior..)
            you can use self.values=None, to disable that error message.
        You can del self.values to re-enable that error message.
            (del self.values will do nothing if self.values doesn't exist, as opposed to crashing.)

        Settings self.values will also reset self.v to None afterwards.
        '''
        try:
            return self._values
        except AttributeError:
            name, plural = self.name, self.plural
            errmsg = (f"{plural} hasn't been set yet!\n(self={object.__repr__(self)})\n"
                      f"You can use: {plural}=None, to disable this error, "
                      f"but it will also disable the 'smart' behavior of setting/getting self.{name}.\n"
                      f"(Note: that 'smart' behavior may have triggered this error message, "
                      f"by trying to get self.{plural}.)\nSee help(type(self).{name}) for more info.\n"
                      f"You can also use: del self.{plural}, to re-enable this error.")
            raise AttributeError(errmsg) from None
    @values.setter
    def values(self, val):
        is_new = (val is not getattr(self, '_values', None))
        self._values = val
        if is_new:
            self.v = None
    @values.deleter
    def values(self):
        try:
            del self._values
        except AttributeError:
            pass  # self.values doesn't exist, so there's nothing to delete.

    loading = simple_property('_loading', default=False,
            doc='''whether this dimension is currently being loaded across, via self.load_across().''')

    # # # METHODS # # #
    @_dim_method('current_n_{dim}')
    def current_n(self):
        '''return number of values currently represented by self.v
        0 if None, else len(self.v) if possible, else 1.
        '''
        v = self.v
        if v is None:
            return 0
        try:
            return len(v)
        except TypeError:
            return 1

    @_dim_method('{dim}_is_iterable')
    def is_iterable(self, *, min_length=None):
        '''return whether self.v represents multiple values.
        False if self.v represents only one value.

        if min_length is provided, return bool((result) and (len(self.v) >= min_length)).
        '''
        return is_iterable_dim(self.v, min_length=min_length)

    @_dim_method('{dim}_list')
    def list(self):
        '''return self.v, but guaranteed to be a list.
        if self.v is iterable, return self.v, unchanged;
        else, return [self.v].
        '''
        v = self.v
        if self.is_iterable():
            return v
        else:
            return [v]

    @_dim_method('iter_{dims}')
    def iter_values(self, *, restore=True, enumerate=False):
        '''iterate through self.values. sets AND yields self.v, for each value in self.values.
        
        restore: bool, default True
            whether to restore original self.v after iteration.
        enumerate: bool, default False
            whether to yield indices too, i.e. (i, v) instead of just v.
            if True, indices correspond to position of value in self.values.
        '''
        with self.maintaining(v=restore):
            for i, val in builtins.enumerate(self.values):
                self.v = val
                if enumerate:
                    yield i, val
                else:
                    yield val

    @_dim_method('iter_{dim}')
    def iter(self, *, restore=True, enumerate=False):
        '''iterate through self.v. sets AND yields self.v, for each value in self.v when iteration began.
        
        Note the difference between this method and iter_values;
            this method uses self.v when called, to determine which values to iterate through,
            while iter_values instead uses self.values to determine which values to iterate through.

        if self.v is not iterable, yield self.v once and do not change its value.

        restore: bool, default True
            whether to restore original self.v after iteration.
        enumerate: bool, default False
            whether to yield indices too, i.e. (i, v) instead of just v.
            if True, indices correspond to position of value in the original self.v (not necessarily self.values!)
        '''
        if not self.is_iterable():
            if enumerate:
                yield 0, self.v
            else:
                yield self.v
            return
        # else, self.v is iterable.
        original = self.v
        with self.maintaining(v=restore):
            for i, val in builtins.enumerate(original):
                self.v = val
                if enumerate:
                    yield i, val
                else:
                    yield val

    @_dim_method('enumerate_{dims}')
    def enumerate_values(self, *, restore=True):
        '''iterate through self.values, seting self.v and yielding (i, v) for each v in self.values.
        Equivalent to self.iter_values(restore=restore, enumerate=True).
        '''
        return self.iter_values(restore=restore, enumerate=True)

    @_dim_method('enumerate_{dim}')
    def enumerate(self, *, restore=True):
        '''iterate through self.v. sets self.v and yields (i, v), for each value in self.v when iteration began.
        i corresponds to position of value in original self.v (not necessarily self.values).
        Equivalent to self.iter(restore=restore, enumerate=True).
        '''
        return self.iter(restore=restore, enumerate=True)

    @_dim_method('iter_{dims}_partition')
    def iter_partition(self, partition, *, restore=True):
        '''iterate through self.values. setting self.v=vlist, yielding (partkey, self.v) from partition.items()
        partition: dict-like of {partkey: vlist}
            use to partition values into groups.
            Note: iter_partition doesn't return vlist directly, but instead sets self.v = vlist,
                then returns self.v. If vlist is a list of v-specifiers, the actual v will be returned.
                E.g. if vlist = [0,3,4], result has self.values.get([0,3,4]), not [0,3,4].
        restore: bool, default True
            whether to restore original self.v after iteration.

        E.g. partition = {'a': ['v0', 'v1', 'v4'], 'b': ['v2', 'v3'], 'c': ['v5']}
            --> yields ('a', ['v0', 'v1', 'v4']), then ('b', ['v2', 'v3']), then ('c', ['v5']),
              setting self.v=['v0', 'v1', 'v4'], then self.v=['v2', 'v3'], then self.v=['v5'].
        '''
        with self.maintaining(v=restore):
            for partkey, partvals in partition.items():
                self.v = partvals
                yield partkey, self.v

    @_dim_method('assign_{dim}_coord')
    def assign_coord(self, array, value=UNSET, *, overwrite=None, expand_if_iterable=False):
        '''returns array.assign_coords(dict({cls.name}=self.v)).
        if not self.is_iterable(), raise DimensionalityError, unless expand_if_iterable=True.

        value: UNSET, or value to use instead of self.v
            if provided, temporarily set self.v = value, then restore it afterwards.
            This means value can be "shorthand" for actual values,
                e.g. value = 0 --> self.v = self.values.get(0).
        overwrite: None or bool
            whether to overwrite an existing value for this coord in array.
            (note - array will never be altered here; only the result might be altered.)
            If this coord already in array.coords, behavior depends on overwrite:
                None --> crash with DimensionKeyError.
                True --> overwrite this coord.
                False --> return array, unchanged.
        expand_if_iterable: bool
            whether to expand_dims if self.is_iterable(),
            e.g. array.expand_dims(dict({cls.name}=self.v))
        '''
        if value is not UNSET:
            with self.using(v=value):
                return self.assign_coord(array, overwrite=overwrite)
        # else, value is UNSET.
        if (not expand_if_iterable) and self.is_iterable():
            errmsg = f"assign_coord requires non-iterable self.v"
            raise self.dimension_value_error(errmsg)
        kw_pass = dict(overwrite=overwrite, expand_if_iterable=expand_if_iterable)
        return xarray_assign(array, coords={self.name: self.v}, **kw_pass)

    @_dim_method('assign_{dim}_along')
    def assign_coord_along(self, array, dim, value=UNSET, **kw_assign_coords):
        '''assign value as coords along the indicated dimension.
        Equivalent to array.assign_coords({self.name: (dim, value)})
        requires self.is_iterable() else raises DimensionValueError.

        value: UNSET, or value to use instead of self.v
            if provided, temporarily set self.v = value, then restore it afterwards.
        '''
        if value is not UNSET:
            with self.using(v=value):
                return self.assign_coord_along(array, dim)
        # else, value is UNSET.
        if not self.is_iterable():
            errmsg = f"assign_coord_along requires iterable self.v"
            raise self.dimension_value_error(errmsg)
        v = self.v
        if isinstance(v, tuple):
            v = list(v)  # for some reason, xarray.assign_coords doesn't like tuples.
        return array.assign_coords({self.name: (dim, v)}, **kw_assign_coords)

    @classmethod
    @_dim_method('take_{dims}')
    def take_along(cls, array, at=None, *, i=None, drop_labels=False,
                   as_dict=False, item=False, **kw__take_along_dimension):
        '''return a list of the array value for each {cls.name} in array.

        array: xarray.DataArray
            array to take from.
            E.g., array with coords fluid ['e', 'H+', 'C+'], cls.name=='fluid'
                --> take_along(array) gives [array value for 'e', value for 'H+', value for 'C+'].
        at: None or list-like of values in this dimension
            take at these values. None --> use values from array.coords[this dimension]
        i: None or indices
            (if provided) take only at these indices; use isel.
        drop_labels: bool, default False
            whether to remove this dimension from arr.coords for arr in result.
        as_dict: bool, default False
            if True, return dict of {dim value: array value at this dim} instead of list.
        item: bool
            if True, convert arrays to single values via array.item().
        additional kwargs go to tools.take_along_dimension
        '''
        kw = dict(at=at, i=i, drop_labels=drop_labels, as_dict=as_dict, item=item, **kw__take_along_dimension)
        return take_along_dimension(cls.name, array=array, **kw)

    @_dim_method('take_{dim}')
    def take(self, array, value=UNSET, *, drop_labels=False,
             as_dict=False, squeeze=True, item=False, **kw__take_along):
        '''take the array at array.{cls.name} == self.v (or at value, if provided).
        if self.is_iterable(), return [array at array.{cls.name}==val for val in self.v].
        otherwise, return a single array: array at array.{cls.name}==self.v.

        array: xarray.DataArray
            array to take from.
            E.g., self.v=='e', cls.name=='fluid', array has coord 'fluid' with one value equal to 'e'
                --> take(array) gives array value at fluid=='e'.
        value: UNSET or value
            if provided, temporarily set self.v = value, then restore it afterwards.
            This means value can be "shorthand" for actual values,
                e.g. value = 0 --> self.v = self.values.get(0)
                    value = None --> self.v = None, so self.v gives self.values.
                    See help(type(self).v) for more info.
        drop_labels: bool, default False
            whether to remove this dimension from arr.coords for arr in result.
        as_dict: bool, default False
            if True, return dict of {dim value: array value at this dim} instead.
        squeeze: bool, default True
            if False, always return list (with length==1 if not self.is_iterable()),
            unless as_dict=True. (Ignore squeeze if as_dict.)
        item: bool
            if True, convert arrays to single values via array.item().
        '''
        kw__take_along.update(drop_labels=drop_labels, as_dict=as_dict, item=item)
        if value is not UNSET:
            with self.using(v=value):
                return self.take(array, squeeze=squeeze, **kw__take_along)
        # else, value is UNSET.
        if self.is_iterable():
            return self.take_along(array, at=self.v, **kw__take_along)
        else:
            result = self.take_along(array, at=[self.v], **kw__take_along)
            if as_dict:
                return result
            elif squeeze:
                return result[0]
            else:
                return result

    @classmethod
    @_dim_method('join_{dims}')
    def join_along(self, arrays, labels=None, **kw__join_along_dimension):
        '''return arrays joined along the {cls.name} dimension.
        if labels is provided, set result[{cls.name}] = labels.
        '''
        if len(arrays) == 0:
            raise DimensionError(f"expected at least one array, but got {len(arrays)} arrays.")
        return join_along_dimension(self.name, arrays=arrays, labels=labels, **kw__join_along_dimension)

    @_dim_method('_as_single_{dim}')
    def _as_single(self, val=None):
        '''return the single value at self.v, or self.v corresponding to val if provided.
        If result is not a single value, raise DimensionValueError.
        '''
        if val is not None:
            with self.using(v=val):
                return self._as_single(val=None)
        v = self.v
        if self.is_iterable():
            errmsg = f"expected single value, but got {type(self.v).__name__} with length={len(v)}"
            raise self.dimension_value_error(errmsg)
        else:
            return v

    @_dim_method('_get_first_{dim}')
    def _get_first(self, val=None):
        '''return first value from self.v, or self.v corresponding to val if provided.
        If iterable, return self.v[0], else return self.v.
        '''
        if val is not None:
            with self.using(v=val):
                return self._as_single(val=None)
        return self.v[0] if self.is_iterable() else self.v

    @_dim_method('{dim}_type')
    def get_type(self, *, check_all=False):
        '''get type of single value in self.v.
        check_all: bool, default False
            if True, check all values in self.v to ensure they are all the same type.
        '''
        result = None
        for val in self.iter():
            if result is None:
                result = type(val)
                if not check_all:
                    break
            elif type(val) != result:
                errmsg = f"cannot get_single_type when types don't all match. Found {result} and {val}."
                raise self.dimension_value_error(errmsg)
        return result

    # # # DISPLAY # # #
    def __repr__(self):
        '''return repr for an instance of this class.'''
        contents = []
        v = self.v
        try:
            rep = v._short_repr()
        except AttributeError:
            rep = repr(v)
        contents.append(f'v={rep}')
        values = self.values
        try:
            rep = values._short_repr()
        except AttributeError:
            rep = repr(values)
        contents.append(f'values={rep}')
        return f'{type(self).__name__}({", ".join(contents)})'



"""
File Purpose: UnitsHaver
"""
import xarray as xr

from .units_manager import UnitsManager
from ..dimensions import BehaviorHaver
from ..tools import alias, alias_child

class UnitsHaver(BehaviorHaver):
    '''has some unit-related methods. self.u manages the actual units.

    u: None or UnitsManager
        manages the units.
        if not None, set self.u = u
    units: None or str
        the current default unit system for outputs. E.g. 'si' or 'raw'.
        if not None, set self.units = units.

    self.units aliases to self.u.units.
        i.e., setting self.units sets self.u.units; getting self.units gets self.u.units.
    '''
    def __init__(self, u=None, units=None, **kw_super):
        if u is not None: self.u = u
        if units is not None: self.units = units
        super().__init__(**kw_super)

    @property
    def u(self):
        '''the current units manager for self.
        if not yet initialized, getting self.u will create (and store) a new UnitsManager().
        '''
        try:
            return self._u
        except AttributeError:
            self._u = UnitsManager()
            return self._u
    @u.setter
    def u(self, value):
        self._u = value

    cls_behavior_attrs.register('units')
    units = alias_child('u', 'units',
            doc='''the current unit system for self. E.g., 'si'. (this is an alias to self.u.units)''')

    def record_units(self, array):
        '''return array.assign_attrs(self.units_meta()).
        if array is not an xarray.DataArray, convert it first.
        '''
        if not isinstance(array, xr.DataArray):
            array = xr.DataArray(array)
        return array.assign_attrs(self.units_meta())

    def units_meta(self):
        '''return dict(units=self.units).'''
        return dict(units=self.units)


class CoordsUnitsHaver(UnitsHaver):
    '''UnitsHaver which allows to set separate units systems for coordinates and data.
    Expects coordinate-units-getting-calls to supply alt='coords' to self.u,
        otherwise will default to coords units being the same as self units.
    '''
    cls_behavior_attrs.register('coords_units', default=None)
    @property
    def coords_units(self):
        '''None or str telling the current unit system for coordinates, e.g. 'si'.
        None --> "coords_units should match self.units".

        Value stored at self.u.alts.get('coords', None)

        See also: coords_units_explicit
        '''
        return self.u.alts.get('coords', None)
    @coords_units.setter
    def coords_units(self, value):
        self.u.alts['coords'] = value
    @coords_units.deleter
    def coords_units(self):
        self.u.alts.pop('coords', None)  # None -> it's okay if coords_units didn't exist yet

    coords_units_explicit = alias('coords_units',
        doc='''str telling the current unit system for coordinates, e.g. 'si'.
        alias to self.coords_units if not None, else self.units.
        equivalent: self.u.alts.get('coords', self.units)''')
    @coords_units_explicit.getter
    def coords_units_explicit(self):
        result = self.coords_units
        return self.units if result is None else result

    def units_meta(self):
        '''return dict(units=self.units). Also include coords_units if different.'''
        result = super().units_meta()
        if self.coords_units is not None:
            result['coords_units'] = self.coords_units_explicit
        return result



"""
File Purpose: DimensionValue, DimensionValueList, UniqueDimensionValue
"""

from ...errors import (
    DimensionValueError, DimensionKeyError, DimensionalityError,
    InputError, InputMissingError,
)
from ...tools import (
    elementwise_property, alias,
    repr_simple,
    Sentinel, UNSET, ATTR_UNSET,
    is_iterable,
    is_integer, interprets_fractional_indexing,
    XarrayIoSerializable,
)
from ...defaults import DEFAULTS


''' --------------------- DimensionValue --------------------- '''

class DimensionValue(XarrayIoSerializable):
    '''value of a dimension. E.g., a single Fluid or a single Snap.
    Also may know how to be converted to str and int.

    s: string or None
        str(self) --> s. if None, cannot convert to str.
    i: nonnegative int, or None
        int(self) --> i. if None, cannot convert to int.
        Must be nonnegative; negative ints are reserved for "index from end of list".

    CAUTION: intended to be immutable (but not strictly enforced)!
        changing the values of s or i after initialization might cause unexpected behavior.
        E.g., DimensionValueList.get utilizes caching and expects s & i to not change.

    testing equality shows that DimensionValue == s or i,
        e.g. val1=DimensionValue('H', 0) --> val1 == 'H' and val1 == 0.
        However when comparing to another DimensionValue, s and i must both be equal,
            e.g. val2=DimensionValue('C', 0) --> val2 != val1.
            (result also depends on DimensionValue types, if subclasses are involved.)
    '''
    # attrs "defining" this DimensionValue, & which can be provided as kwargs in __init__
    _kw_def = {'s', 'i'}

    _kw_eq = alias('_kw_def', doc='''attributes used when testing equality. Default: alias to _kw_def.''')
    _kw_copy = alias('_kw_def', doc='''attributes used as defaults in copy(). Default: alias to _kw_def.''')

    def __init__(self, s=None, i=None):
        self.s = s
        self.i = i
        if (i is not None) and (i < 0):
            raise DimensionValueError(f"negative i not allowed, but got i={i}.")
    def __str__(self):
        return f'{type(self).__name__}_{self.i}' if self.s is None else self.s
    def __int__(self):
        return self.i

    def __eq__(self, v):
        '''return self==v. Equal if:
            - v is an instance of type(self), and
                all kw from both self._kw_eq and v._kw_eq have the same values.
                    (any kw missing attrs will be treated as UNSET.)
                For DimensionValue, just compares self.s==v.s and self.i==v.i.
                    Subclasses may add more kwargs to compare by altering _kw_def.
            - or, self.s == v or self.i == v, and v is not another DimensionValue
            - or, v is a DimensionSingleValueSpecifier and v.checker(self) gives True.
        Not equal if v is a DimensionValue but not an instance of type(self),
            e.g. class Fluid(DimensionValue):...; class Snap(DimensionValue):...;
                Fluid('H', 0) != Snap('H', 0).
        '''
        if v is self:
            return True   # an object is equal to itself
        elif isinstance(v, DimensionSingleValueSpecifier):
            return v.checker(self)
        elif isinstance(v, type(self)):
            kw_compare = self._kw_eq.union(v._kw_eq)
            for key in kw_compare:
                if getattr(self, key, UNSET) != getattr(v, key, UNSET):
                    return False
            return True
        elif isinstance(v, DimensionValue):
            if DEFAULTS.DEBUG >= 4:
                print('DimensionValue comparison of incompatible subclasses: ',
                      type(self), type(v), '--- returning False.')
            return False
        else:
            return (self.s == v) or (self.i == v)

    def equal_except_i(self, v):
        '''returns whether self == v (another DimensionValue), ignoring i.
        equal if v is an instance of type(self) and
            all kw from both self._kw_eq and v._kw_eq have the same values.
            (any kw missing attrs will be treated as UNSET.)
            For DimensionValue: self.s==v.s and self.i==v.i. Subclasses may add more.
        '''
        if not isinstance(v, type(self)):
            return False
        kw_compare = self._kw_eq.union(v._kw_eq)
        for key in kw_compare:
            if key == 'i': continue
            if getattr(self, key, UNSET) != getattr(v, key, UNSET):
                return False
        return True

    _hash_tuple_convert_len_limit = 100  # if hashing, convert shorter non-hashable iterables to tuples first.

    def __hash__(self):
        result = [getattr(self, k, ATTR_UNSET) for k in self._kw_eq]
        try:
            result = hash((type(self), *result))
        except TypeError:  # maybe one or more of the things in result were not hashable.
            for i, r in enumerate(result):
                try:  # [TODO][EFF] if this is ever a computational bottleneck, be more efficient
                    hash(r)
                except TypeError:
                    if is_iterable(r):
                        if len(r) > self._hash_tuple_convert_len_limit:
                            errmsg = (f"cannot hash {type(self).__name__} with unhashable iterable {list(self._kw_eq)[i]}"
                                      f" with length > self._hash_tuple_convert_len_limit (= {self._hash_tuple_convert_len_limit}).")
                            raise TypeError(errmsg)
                        result[i] = tuple(r)
                    else:
                        raise
            result = hash((type(self), *result))
        return result

    def copy(self, **kw_init):
        '''return a copy of self. Can provide new kwargs here to override old values in result.
        E.g. self.copy(i=7) makes a copy of self but with i=7 instead of self.i.
        '''
        for key in self._kw_copy:
            kw_init.setdefault(key, getattr(self, key))
        return type(self)(**kw_init)

    def with_i(self, i):
        '''return copy of self with i=i, or self if self.i==i already.'''
        if self.i == i:
            return self
        return self.copy(i=i)

    def to_dict(self):
        '''return dictionary of info about self. Attribute values for keys in self._kw_def.
        e.g. if _kw_def={'s', 'i'}: result = {'s': self.s, 'i': self.i}
        '''
        return {k: getattr(self, k) for k in self._kw_def}

    def lookup_dict(self):
        '''returns dict for looking up self within a DimensionValueList, given int, str, or self.
        (used by DimensionValueList.lookup_dict)
        '''
        result = dict()
        result[self] = self
        try:
            result[str(self)] = self
        except TypeError:
            pass
        try:
            result[int(self)] = self
        except TypeError:
            pass
        return result

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}({self.s!r}, {self.i!r})'

    def _repr_simple(self):
        '''return simple repr for self: Classname(s, i).
        if s is None, use Classname(i) instead.
        if i is None, use Classname(s, i=None) instead.
        Called when using PlasmaCalcs.tools.repr_simple.
        '''
        if self.s is None:
            return f'{type(self).__name__}({self.i!r})'
        elif self.i is None:
            return f'{type(self).__name__}({self.s!r}, i=None)'
        else:
            return f'{type(self).__name__}({self.s!r}, {self.i!r})'

    # ndim = 0 is useful for achieving desired behavior,
    #  of being treated as "a single value of a dimension",
    #  while still allowing subclasses to implement __iter__.
    # ndim = 0 is checked by:
    #  - DimensionHaver, when checking dim_is_iterable() for an instance of this class.
    #  - xarray, when attempting to use an instance of this class as a coordinate.
    ndim = 0

    # size = 1 is also useful.
    size = 1

    # sometimes it's convenient to have an ordering for the values...
    def __lt__(self, other):
        if isinstance(other, type(self)): return (self.i, self.s) < (other.i, other.s)
        else: return NotImplemented
    def __le__(self, other):
        if isinstance(other, type(self)): return (self.i, self.s) <= (other.i, other.s)
        else: return NotImplemented
    def __gt__(self, other):
        if isinstance(other, type(self)): return (self.i, self.s) > (other.i, other.s)
        else: return NotImplemented
    def __ge__(self, other):
        if isinstance(other, type(self)): return (self.i, self.s) >= (other.i, other.s)
        else: return NotImplemented


''' --------------------- DimensionValueList --------------------- '''

class SliceableList(list):
    '''list that returns a new instance of type(self) when sliced.
    Also permits slicing by an iterable (take one element for each index in the iterable).
    '''
    def _new(self, *args, **kw):
        '''return new instance of type(self), with these args and kw.'''
        return type(self)(*args, **kw)

    def __getitem__(self, ii, **kw_new):
        '''return self[ii], giving element from self if int(ii) possible, else new instance of type(self).'''
        try:
            i = int(ii)
        except TypeError:
            pass  # handled after 'else' block
        else:
            return super().__getitem__(i)
        if isinstance(ii, slice):
            return self._new(super().__getitem__(ii), **kw_new)
        else:
            return self._new([super(SliceableList, self).__getitem__(i) for i in ii], **kw_new)

    size = property(lambda self: len(self), '''alias to len(self).''')

    def inserted(self, i, element, **kw_new):
        '''return copy of self but with element inserted at i. kwargs passed to new self.__init__.'''
        elements = list(self)
        elements.insert(i, element)
        return self._new(elements, **kw_new)

    # # # DISPLAY # # #
    def _display_contents(self, f, *, nmax=3):
        '''return str to display self's contents, using f(element) for each element.
        if len(self) > nmax, instead include len(self), self[0], and self[-1], only.
        '''
        if len(self) <= nmax:
            contents = ', '.join(f(el) for el in self)
        else:
            contents = f'len={len(self)}; {f(self[0])}, ..., {f(self[-1])}'
        return contents

    def __repr__(self):
        contents = self._display_contents(repr, nmax=3)
        return f'{type(self).__name__}({contents})'

    def _repr_simple(self):
        '''return simple repr for self; includes type, len, and a repr_simple for a few elements.
        if less than 4 elements, just uses repr_simple for the elements (doesn't include len).
        '''
        contents = self._display_contents(repr_simple, nmax=3)
        return f'{type(self).__name__}({contents})'
    
    def __str__(self):
        contents = self._display_contents(str, nmax=3)
        return f'{type(self).__name__}({contents})'

    def _short_repr(self):
        '''return a shorter repr of self; just includes type & len.'''
        return f'{type(self).__name__}(with length {len(self)})'


def string_int_lookup(dict_):
    '''return dict_ but with DimensionValue, str, and int keys. dict_ keys should be DimensionValue objects.
    if any keys can't be converted to str or int, skip converting those keys to that type.
    '''
    # [TODO] remove this. Already refactored to use lookup_dict from DimensionValue.
    #  (string_int_lookup is used elsewhere though, need to investigate further.)
    result = dict()
    for key, val in dict_.items():
        result[key] = val
        try:
            result[str(key)] = val
        except TypeError:
            pass
        try:
            result[int(key)] = val
        except TypeError:
            pass
    return result

def string_int_index(list_):
    '''return dict mapping DimensionValue, str, and int keys to the indices of corresponding DimensionValue values.'''
    return {k: i for i, val in enumerate(list_) for k in (val, str(val), int(val))}

class DimensionValueList(SliceableList):
    '''SliceableList of DimensionValue elements. Also provides get() and lookup_dict() methods,
    which allow to get values based on int or string (or slice, or tuple or int or strings).
    Those methods use caching (assume int(element) & str(element) never changes for any element).

    istart: None or int
        if provided, reindex elements with new i values, starting from istart.
        E.g. istart=0 --> result.i == 0, 1, 2, ....
        elements which already have correct i will remain unchanged.
    '''
    _dimension_key_error = DimensionKeyError
    value_type = DimensionValue

    ndim = 1  # primary use: to distinguish between this and DimensionValue.

    s = elementwise_property('s')

    # # # CREATION OPTIONS # # #
    def __init__(self, *args_list, istart=None, **kw_list):
        if istart is None:
            super().__init__(*args_list, **kw_list)
        else:
            ll = list(*args_list, **kw_list)
            for i, el in enumerate(ll):
                ll[i] = el.with_i(i=istart+i)
            super().__init__(ll)

    @classmethod
    def from_strings(cls, strings):
        '''return cls instance from iterable of strings. (i will be determined automatically.)
        Equivalent to cls(cls.value_type(s, i) for i, s in enumerate(strings)).
        '''
        return cls(cls.value_type(s, i) for i, s in enumerate(strings))

    @classmethod
    def from_dicts(cls, dicts, *, istart=UNSET, **kw_value_init):
        '''return cls instance from iterable of dicts

        values will all be DimensionValue objects, with type == cls.value_type.

        dicts: iterable of dicts
            each one corresponds to one (cls.value_type) object in result.
        istart: UNSET, None, or int
            mode for inferring i values if needed.
            UNSET --> dicts[k] uses i=dicts[k]['i'] if it exists, else i=k.
            None --> use i from dict if it exists, else None.
            int --> dicts[k] uses i=istart+k, ignoring any 'i' provided in dict.
        additional kwargs will be passed to cls.value_type.__init__
            (in case of conflict with any dicts values, use dicts values).
        '''
        if istart is UNSET:
            i_vals = [d.get('i', k) for k, d in enumerate(dicts)]
            dicts = [{**d, 'i': i} for i, d in zip(i_vals, dicts)]
        elif istart is None:
            dicts = [d if 'i' in d else {**d, 'i': None} for d in dicts]
        else:
            dicts = [{**d, 'i': istart+k} for k, d in enumerate(dicts)]
        return cls(cls.value_type(**kw_value_init, **d) for d in dicts)

    @classmethod
    def from_dict(cls, int_to_element):
        '''return DimensionValueList from dict of {i: element}.
        See also: cls.from_dicts()
        '''
        result = cls(el for i, el in sorted(int_to_element.items()))
        for i, el in enumerate(result):
            if (i != el.i) and (el.i is not None):
                raise ValueError(f"expected i and element.i to match. i={i}, element.i={el.i}. (element={el})")
        return result

    @classmethod
    def from_array(cls, array):
        '''return DimensionValueList from 0D or 1D array.
        array values should be DimensionValue objects, strings, or ints.
            DimensionValue objects --> use as-is
            strings --> use cls.value_type(s, i=i) for i in range(len(array))
            ints --> use cls.value_type(i=i) for i in array
        '''
        if array.ndim == 0:
            values = [array.item()]
        elif array.ndim == 1:
            values = array.values
        else:
            errmsg = f'{cls.__name__}.from_array expects array.ndim=0 or 1; got ndim={array.ndim}'
            raise DimensionalityError(errmsg)
        result = []
        for i, v in enumerate(values):
            if isinstance(v, cls.value_type):
                result.append(v)
            elif isinstance(v, str):
                result.append(cls.value_type(v, i=i))
            elif is_integer(v):
                result.append(cls.value_type(i=v))
            else:
                errmsg = (f'{cls.__name__}.from_array got unexpected value type at index {i}.\n'
                            f'Expected {cls.value_type.__name__}, str, or int; got value={v!r}.')
                raise InputError(errmsg)
        return cls(result)

    @classmethod
    def unique_from(cls, elements, *elements_as_args, istart=None):
        '''return DimensionValueList of unique elements from elements.
        equality checked via element.equal_except_i, i.e. ignore index.

        elements: iterable of DimensionValue of cls.value_type.
            or, single DimensionValue, in which case treat it as first arg,
            and can provide more elements as additional args.
        istart: None or int. int --> use result[k].with_i(istart + k) for all k.
        '''
        # bookkeeping
        if elements_as_args:
            elements = (elements,) + elements_as_args
        else:
            ndim = getattr(elements, 'ndim', None)
            if ndim is None:
                ndim = 1 if is_iterable(elements) else 0
            if ndim == 0:
                elements = (elements,)
        # getting result
        result = []
        for e in elements:
            for r in result:
                if e.equal_except_i(r):
                    break
            else:  # did not break, i.e. element not found
                result.append(e)
        return cls(result, istart=istart)

    def unique(self, *, istart=None):
        '''return DimensionValueList of unique elements from self.
        equality checked via element.equal_except_i, i.e. ignore index.
        (if all elements are unique, return self, else return new DimensionValueList.)
        istart: None or int. int --> use result[k].with_i(istart + k) for all k.
        '''
        result = []
        for e in self:
            for r in result:
                if e.equal_except_i(r):
                    break
            else:  # did not break, i.e. element not found
                result.append(e)
        cls = type(self)
        if len(result) == len(self):
            result = self.with_istart(istart)
        else:
            result = cls(result, istart=istart)
        return result

    # # # REINDEXING / MANIPULATING # # #
    reindexed = alias('with_istart')

    def with_istart(self, istart=0):
        '''return new DimensionValueList with elements reindexed, starting from istart.
        elements with incorrect i will be replaced with copies.
        if istart is None, just return self, unchanged.
        '''
        if istart is None:
            return self
        return type(self)(el.with_i(i=istart+i) for i, el in enumerate(self))

    # # # LOOKUPS / GETTING ELEMENTS # # #
    def lookup_dict(self):
        '''return dict for looking up elements given int or str (or element) in self.
        uses caching; assumes int(element) & str(element) never changes for any element.
        '''
        # [TODO] Doesn't work when changing distribution names using fluids[N].s = Name.
        try:
            lookup = self._lookup
        except AttributeError:
            lookup = dict()
            for element in self:
                lookup.update(element.lookup_dict())
            self._lookup = lookup
        return lookup

    def lookup(self, key, *, unique_missing_ok=False, handle_special=True):
        '''return element given int, str, or element in self. raise DimensionKeyError if key not in self.
        unique_missing_ok: bool
            whether it is okay for self to be missing key if key is a UniqueDimensionValue.
            if True and key is a UniqueDimensionValue, return key instead of raising error.
        handle_special: bool
            whether to handle DimensionSpecialValueSpecifier in special way,
                i.e. if key is a DimensionSpecialValueSpecifier, return key.getter(self).
        '''
        if unique_missing_ok and isinstance(key, UniqueDimensionValue):
            return key
        if handle_special and isinstance(key, DimensionSpecialValueSpecifier):
            return key.getter(self)
        try:
            return self.lookup_dict()[key]
        except KeyError:
            raise self._dimension_key_error(key) from None

    def get(self, key):
        '''return element(s) corresponding to key.
        key: None, int, str, element, DimensionSpecialValueSpecifier, range, slice, or tuple/list
            Some keys will look up and return corresponding element(s):
                nonnegative int, str, or element --> return corresponding element.
                DimensionSpecialValueSpecifier spec --> return spec.getter(self).
                tuple --> return tuple(self.get(k) for k in key)
                list --> return list(self.get(k) for k in key)
                range --> return type(self)(self.get(k) for k in key).
            Other keys will index self using the usual list-like indexing rules:
                None --> return self, unchanged.
                negative int --> return self[key], i.e. the key'th element, counting from end.
                slice --> return self[key], i.e. apply this slice to self.
                        Note that the result will have the same type as self.
                        Note that this supports interprets_fractional_indexing,
                          e.g. slice(0.3, 0.7) will slice(len(self) * 0.3, len(self) * 0.7).
            if any element in a tuple or list is not in self,
                keep it unchanged if it is a UniqueDimensionValue,
                else raise DimensionKeyError.
        Note that the result will always be single element, a tuple, a list, or a DimensionValueList.
        '''
        if key is None:
            return self
        elif isinstance(key, DimensionSpecialValueSpecifier):
            return key.getter(self)
        # index self like it is a list:
        elif isinstance(key, slice):
            key = interprets_fractional_indexing(key, len(self))
            return self[key]
        elif is_integer(key) and key < 0:
            return self[key]
        # index by looking up key(s) in self:
        elif isinstance(key, (tuple, list)):
            return type(key)(self.lookup(k, unique_missing_ok=True) for k in key)
        elif isinstance(key, range):
            return type(self)(self.lookup(k) for k in key)
        else:
            return self.lookup(key, unique_missing_ok=False)

    def indices(self, key, *, return_get=False, to_list=False):
        '''return indices of element(s) corresponding to key, during self.get(key).
        to_list: bool.
            True --> return [index] instead of index, if self.get(key) is a single DimensionValue.
        return_get: bool
            True --> return (indices, self.get(key)) instead of just indices.
        result will have same length as self.get(key); self[result[i]] is self.get(key)[i].
        '''
        id2index = {id(el): i for i, el in enumerate(self)}
        if len(id2index) != len(self):
            raise NotImplementedError(f'{type(self).__name__}.indices() with non-unique elements.')
        got = self.get(key)
        if isinstance(got, DimensionValue):
            if to_list:
                got = [got]
            else:
                return id2index[id(got)]
        indices = [id2index[id(g)] for g in got]
        return (indices, got) if return_get else indices

    def without_i(self, i, **kw_init):
        '''return a copy of self but dropping elements currently at these indices.
        i: int or iterable of int. indices to remove.
        kw_init are passed to __init__ when making the copy of self.
        '''
        if not is_iterable(i):
            i = [i]
        ikeep = sorted(set(range(len(self))) - set(i))
        return self.__getitem__(ikeep, **kw_init)

    def popped(self, key, *, return_indices=False, **kw_init):
        '''return (self.get(key), self.without_i(indices corresponding to key)).
        see help(self.get) for details on allowed keys.

        return_indices: bool
            if True, also return the indices (as a list) of the popped element(s),
            i.e. return (indices, popped element(s), self.without_i(...), self with elements removed).
            index will be None if no elements were removed.
        kw_init are used when making the copy of self. E.g. use istart=0 to make the copy start at 0.
        '''
        indices, pop = self.indices(key, return_get=True, to_list=True)
        newself = self.without_i(indices, **kw_init)
        return (index, pop, newself) if return_index else (pop, newself)

    # # # SERIALIZING # # #
    def to_dict(self):
        '''returns a list of dictionaries,
        each dictionary corresponds to each DimensionValue in the DimensionValueList.
        '''
        result_list = []
        for dimension_value in self:
            result_list.append(dimension_value.to_dict())
        return result_list

    def serialize(self):
        '''return serialization of self, into a list of dictionaries.'''
        # [TODO][EFF] use dimension_value.serialize(include_typename=False) for all except the first?
        result_list = []
        for dimension_value in self:
            result_list.append(dimension_value.serialize())
        return result_list


''' --------------------- DimensionSpecialValueSpecifier --------------------- '''

class DimensionSpecialValueSpecifier():
    '''behaves in a special way when used to specify a dimension value:
        dim.v = specifier --> dim.v = specifier.getter(dim.values)
        dlist.get(specifier) --> specifier.getter(dlist)
    (for DimensionSpecialValueSpecifier specifier,
        Dimension dim, and DimensionValueList dlist.)
    
    getter: str or callable
        self specifies to use the dimension value(s): self.getter(dlist)
        callable --> callable of 1 arg; will be passed dlist.
        str --> use dlist.{getter}(). E.g. getter='get_ion' --> use dlist.get_ion().
        internally stored at self.init_getter.
    iseler: None, str, or callable
        if provided, can use self in calls to PlasmaCalcs' xarray_isel and xarray_sel,
            commonly via array.pc.sel(dim=self) or array.pc.isel(dim=self).
        None --> crash with InputError if used in xarray_isel or xarray_sel.
        callable --> callable of 1 arg; will be passed list of coord values for dim,
            (probably a list of DimensionValue objects, maybe converted to DimensionValueList).
            must return index (or indexes) appropriate for xarray.isel.
        str --> use dlist.{iseler}(), where dlist=isel_cls(list of coord values for dim).
            E.g. iseler='i_ions', isel_cls=FluidList --> use FluidList(values).i_ions().
        internally stored at self.init_iseler.
    isel_cls: UNSET, None, or type
        if provided, iseler calls use iseler(isel_cls(list of coord values for dim)),
        else, iseler calls use iseler(list of coord values for dim).
        (None --> set self.isel_cls = None. UNSET --> do not set self.isel_cls.)
    '''
    def __init__(self, getter, *args__None, iseler=None, isel_cls=UNSET, **kw__None):
        self.init_getter = getter
        if isinstance(getter, str):
            self.getter = self._getter_from_str
        else:
            self.getter = getter
        self.init_iseler = iseler
        if isinstance(iseler, str):
            self.iseler = self._iseler_from_str
        else:
            self.iseler = self._iseler_when_none
        if isel_cls is not UNSET:
            self.isel_cls = isel_cls

    def _getter_from_str(self, dlist):
        '''get the value from dlist specified by str self.getter.'''
        return getattr(dlist, self.init_getter)()

    def _iseler_from_str(self, dvals):
        '''get the indexes from dlist specified by str self.iseler.'''
        if getattr(self, 'isel_cls', None) is None:
            raise InputMissingError("coding error. self.isel_cls is required if using str iseler.")
        dlist = self.isel_cls(dvals.values)
        return getattr(dlist, self.init_iseler)()

    def _iseler_when_none(self, dvals):
        '''crash, because self was passed to xarray_isel or xarray_sel but self.iseler is None.'''
        raise InputError(f"{self} has iseler=None, so it cannot be used in xarray_isel or xarray_sel.")

    # # # DISPLAY # # #
    def _getter_as_str(self):
        '''returns str representation of self.getter.
        Use self.init_getter if it exists, else self.getter if it exists, else 'UNKNOWN_GETTER'.
        '''
        return getattr(self, 'init_getter', getattr(self, 'getter', 'UNKNOWN_GETTER'))

    def _iseler_as_str(self):
        '''returns str representation of self.iseler.
        Use self.init_iseler if it exists, else self.iseler if it exists, else 'UNKNOWN_ISELER'.
        '''
        return getattr(self, 'init_iseler', getattr(self, 'iseler', 'UNKNOWN_ISELER'))

    def _repr_contents(self):
        '''contents for repr; used by __repr__.'''
        result = [f'getter={self._getter_as_str}']
        if self.init_iseler is not None:
            result.append(f'iseler={self._iseler_as_str}')
        return result

    def __repr__(self):
        return f'{type(self).__name__}({", ".join(self._repr_contents())})'


class DimensionSingleValueSpecifier(DimensionSpecialValueSpecifier):
    '''DimensionSpecialValueSpecifier which includes checker, to check if single value matches.
    getter gets appropriate single value from list.
    checker checks if value is a match for this specifier.

    e.g. fluids.py defines ELECTRON,
        with getter calling FluidList.get_electron,
        and checker calling Fluid.is_electron().

    getter: str or callable
        self specifies to use the dimension value(s): self.getter(dlist)
        callable --> callable of 1 arg; will be passed dlist.
        str --> use dlist.{getter}(). E.g. getter='get_ion' --> use dlist.get_ion().
        internally stored at self.init_getter.
    checker: str or callable
        when checking value==self (or self==value): use self.checker(value)
        callable --> callable of 1 arg; will be passed value.
        str --> use value.{checker}(). E.g. checker='is_electron' --> use value.is_electron().
        internally stored at self.init_checker.
    iseler: None, str, or callable
        if provided, can use self in calls to PlasmaCalcs' xarray_isel and xarray_sel,
            commonly via array.pc.sel(dim=self) or array.pc.isel(dim=self).
        callable --> callable of 1 arg; will be passed list of coord values for dim,
            (probably a list of DimensionValue objects, maybe converted to DimensionValueList).
            must return index (or indexes) appropriate for xarray.isel.
        str --> use dlist.{iseler}(), where dlist=isel_cls(list of coord values for dim).
            E.g. iseler='i_ions', isel_cls=FluidList --> use FluidList(values).i_ions().
        internally stored at self.init_iseler.
    isel_cls: None or type
        if provided, iseler calls use iseler(isel_cls(list of coord values for dim)),
        else, iseler calls use iseler(list of coord values for dim).
    '''
    def __init__(self, getter, checker, *args_super, **kw_super):
        super().__init__(getter, *args_super, **kw_super)
        self.init_checker = checker
        if isinstance(checker, str):
            self.checker = self._checker_from_str
        else:
            self.checker = checker

    def _checker_from_str(self, value):
        '''check if value is a match for str self.checker.'''
        return getattr(value, self.init_checker)()

    # # # DISPLAY # # #
    def _checker_as_str(self):
        '''returns str representation of self.checker.
        Use self.init_checker if it exists, else self.checker if it exists, else 'UNKNOWN_CHECKER'.
        '''
        return getattr(self, 'init_checker', getattr(self, 'checker', 'UNKNOWN_CHECKER'))

    def _repr_contents(self):
        '''contents for repr; used by __repr__.'''
        result = super()._repr_contents()
        result.append(f'checker={self._checker_as_str}')
        return result


class DimensionSpecialValue(DimensionSingleValueSpecifier):
    '''a single special value for a dimension, which doesn't necessarily appear in '''


class DimensionFractionalIndexer(DimensionSpecialValueSpecifier):
    '''DimensionSpecialValueSpecifier which specifies to do interprets_fractional_indexing.
    
    Example:
        indexer = DimensionFractionalIndexer(0.5)
        snaps.get(indexer) --> gets the value at 0.5*len(snaps).
        DimensionFractionalIndexer([0.2, 0.3])  # values at 0.2*len(dlist) and 0.3*len(dlist).
        DimensionFractionalIndexer(slice(0, None, 0.1))  # values at 0%, 10%, 20%, ..., of len(dlist).
    '''
    def __init__(self, indexer, *args__None, **kw__None):
        self.indexer = indexer

    def _getter_as_str(self):
        '''return str representation of self.indexer.'''
        return self.indexer

    def getter(self, dlist):
        '''get the value(s) from dlist specified by self.indexer, with interprets_fractional_indexing.'''
        indexes = interprets_fractional_indexing(self.indexer, len(dlist))
        return dlist[indexes]


''' --------------------- SpecialDimensionValue --------------------- '''

class SpecialDimensionValue(DimensionValue, DimensionSpecialValueSpecifier):
    '''a special dimension value, not corresponding to a usual value from a DimensionValueList.
    E.g., BifrostScrSnap is a SpecialDimensionValue for the snap corresponding to the scr files.
        Even though BifrostScrSnap instances won't appear in BifrostSnapList,
        we still want to allow bifrost_calculator.snap = BifrostScrSnap(...).
    '''
    def getter(self, dlist):
        '''get the value from dlist specified by self.
        (Here, just returns self, the SpecialDimensionValue)
        '''
        return self


''' --------------------- UniqueDimensionValue --------------------- '''

class UniqueDimensionValue(Sentinel, SpecialDimensionValue):
    '''a unique dimension value, not corresponding to a usual value from a DimensionValueList.
    E.g., INPUT_SNAP (defined in snaps.py) is the UniqueDimensionValue for the snap corresponding to the input deck.

    Cannot provide any args or kwargs to __init__.
    Equality with other dimension values only holds if the other value is a UniqueDimensionValue of the same type.
    str(UniqueDimensionValue) will return the type name.
    int(UniqueDimensionValue) will return None.
    '''
    _kw_def = {'name'}
    # _kw_copy = {'name'}   # might be necessary if subclasses are editing _kw_copy defaults, but not sure...

    def __init__(self, *args__None, **kw__None):
        pass  # do nothing (instead of super().__init__)

    i = None
    name = None  # overwrite parent's 'name' property, if one exists...
    s = alias('name')  # from Sentinel

    def __hash__(self):
        return super().__hash__()  # we need to do this since we define __eq__ here...

    def __eq__(self, other):
        return type(self) == type(other)

    def copy(self, **kw__None):
        '''return self; self is a Sentinel so there is only 1 instance.'''
        return self



"""
File Purpose: fft of xarrays.
"""

import numpy as np
import xarray as xr

from .array_fft import fftN, ifftN
from .fft_dimnames import FFTDimname
from .fft_slices import FFTSlices, _paramdocs_fft_slices
from ..docs_tools import format_docstring
from ..iterables import is_iterable
from ..sentinels import UNSET
from ..xarray_tools import (
    pcAccessor, xarray_promote_dim, xarray_get_dx_along,
)
from ...defaults import DEFAULTS
from ...errors import InputConflictError, InputMissingError


''' --------------------- fft --------------------- '''

@pcAccessor.register('fftN', aliases=['fft'], totype='array')
@format_docstring(**_paramdocs_fft_slices)
def xarray_fftN(array, dim=None, ds=None, *, rad=None, abs=False,
                slices=dict(), keep=UNSET, half=UNSET, step=UNSET, missing_slices=UNSET, **kw_np_fftn):
    '''calculates fft(array) along N dimensions.
    shifts frequencies such that the 0-frequency is in the center.
    replaces result dimensions & coordinates appropriately, to indicate which dims were fft'd.

    dim: None, str, or iterable of strs
        coordinates(s) to take fft over.
        Can be pre-fft or post-fft names. (e.g. 'x', 'freq_x', 'freqrad_x', 'k_x')
        promote_dim(array, coord) for any non-dimension coordinates, as needed.
        None --> equivalent to array.dims
        str --> just this coordinate.
        iterable of strs --> just these coordinates.
    ds: None, number, or dict of {{dim: d}}
        spacing between elements of array (pre-transform), along each dim.
        if number, use the same value for all dims.
        if None, infer via array.coords[dim].diff(dim) for each dim
            (requires evenly-spaced coordinates in dim; spacing checked with np.allclose)
    rad: None or bool
        whether to convert frequencies to "radians" by multiplying them by 2 * pi.
        E.g., for fft in space, rad=False gives 1/wavelength; rad=True gives wavenumber k.
        if None, infer from dim if any post-fft names provided, else default to False.
    abs: bool
        if True, return np.abs(result), instead.
    slices: dict or FFTSlices
        instructions for slicing the final result.
        Can provide {{cname: indexer}} instructing to slice post-fft dimension
            associated with cname, via indexer. (cname can be pre-fft or post-fft name.)
            These understand fractional indexing: can provide a fractional value
            between -1 and 1, to use that fraction of the length along the relevant dimension.
        Can also provide `keep`, `half`, `step`, and/or `missing_slices`, here (Or, as kwargs).
            (raise InputConflictError if any provided in both places and have conflicting values.)
        See those kwargs for more details.
    keep: {keep}
        UNSET --> use None.
    half: {half}
        UNSET --> use None.
    step: {step}
        UNSET --> use None.
    missing_slices: {missing_slices}
        UNSET --> use 'ignore'

    additional kwargs passed to np.fft.fftn.

    returns result of fftn(...), shifted such that the 0-frequency is in the center,
        and with the relevant dimensions renamed as specified.
    '''
    # # parse input # #
    _input_dim = dim  # store input values --> helpful for debugging
    _input_rad = rad
    _input_slices = slices
    # dim
    dims = array.dims if (dim is None) else ([dim] if isinstance(dim, str) else list(dim))
    # freq_dims
    fnames = []
    for dim in dims:
        fname = FFTDimname.implied_from(dim, array.coords, rad=rad)  # (will crash if bad dim or rad)
        fnames.append(fname)
        rad = fname.rad  # <-- all fnames must have compatible rad!
    freq_dims = {fd.pre: fd.post for fd in fnames}   # {pre-fft: post-fft}
    dims = list(freq_dims)  # dims is now the pre-fft names.
    # promote non-dimension coordinates to dimensions, if necessary
    for dim in dims:
        array = xarray_promote_dim(array, dim)
    # ds
    if ds is None:
        ds = {dim: xarray_get_dx_along(array, dim, float_rounding=True) for dim in dims}
    elif isinstance(ds, str):  # raise helpful error message. Maybe caused by typo when inputting args.
        raise TypeError(f'expected None or dict-like ds; got ds={ds!r}')
    elif not is_iterable(ds):
        ds = {dim: ds for dim in dims}
    # slices
    if isinstance(slices, FFTSlices):
        slices = slices.as_kw()
    else:
        slices = slices.copy()  # make a copy, so we can update it without modifying the input.
    kw_slices = dict(keep=keep, half=half, step=step, missing_slices=missing_slices)
    for k, v in kw_slices.items():
        if v is not UNSET:
            if k not in slices:
                slices[k] = v
            elif slices[k] != v:
                raise InputConflictError(f'provided {k!r} both in slices and as kwargs, with conflicting values!')
    slices = FFTSlices(**slices)
    # # convert input to numpy-acceptable format # #
    np_array = array.values
    axes = tuple(array.dims.index(dim) for dim in dims)
    np_ds = tuple(ds[dim] for dim in dims)
    # # take the fft (and handle the fftshift, too) # #
    vals, freqs = fftN(np_array, ds=np_ds, axes=axes, rad=rad, **kw_np_fftn)
    # # converting to xarray # #
    # dims (order matters, must match dims order in vals)
    result_dims = list(array.dims)
    for i, dim in enumerate(result_dims):
        if dim in freq_dims:
            result_dims[i] = freq_dims[dim]
    # coords (update any coordinate info relevant to fft)
    freq_coords = {d: f for d, f in zip(dims, freqs)}    # dict of {original dim name: associated freq values}
    result_coords = array.coords.copy(deep=False)
    for cname, cval in tuple(result_coords.items()):
        if cname in freq_dims:  # a dimension where we applied fft.
            del result_coords[cname]
            fd = freq_dims[cname]
            result_coords[fd] = freq_coords[cname]
        elif any(cd in dims for cd in cval.dims):  # otherwise associated with a dimension where we applied fft.
            del result_coords[cname]  # discard info -- fft'd dims coords don't really map 1-to-1 to input coords.
        else:  # not associated with a dimension where we applied fft.
            pass  # keep the coordinate information.
    # if DEFAULTS.DEBUG:   # if debugging, these lines might help.
    #     return vals, result_dims, result_coords
    attrs = array.attrs.copy() if xr.get_options()['keep_attrs'] else None
    result = xr.DataArray(vals, dims=result_dims, coords=result_coords, attrs=attrs)
    # # slice the result if slices provided # #
    result = slices.applied_to(result, dims=dims)
    # # abs if requested # #
    if abs:
        result = np.abs(result)
    return result


''' --------------------- ifft --------------------- '''

@pcAccessor.register('ifftN', aliases=['ifft'], totype='array')
def xarray_ifftN(array, dim=None, df=None, *, rad=None, pos_dims=None, ds=None, x0=0, **kw_np_ifftn):
    '''calculates ifft(array) along N dimensions.
    shifts positions such that the 0-position is in the center.
    replaces result dimensions & coordinates appropriately, to indicate which dims were ifft'd.

    For convenience, all coordinate names can be pre-fft OR post-fft names,
        e.g. 'x', 'freq_x', 'freqrad_x', or 'k_x'.
        "post-fft" names look like 'freq_dim', 'freqrad_dim',
            or any value in DEFAULTS.FFT_FREQ_RAD_DIMNAMES.values(), e.g. 'k_x'.

    Caution: ifft(fft(arr)) == arr only approximately, due to floating point rounding errors.
        Can at least ensure coordinate alignment by providing ds during ifft(fft(arr), ds=...)

    dim: None, str, or iterable of strs
        coordinates(s) of array to take ifft over.
        promote_dim(array, coord) for any non-dimension coordinates, as needed.
        None --> equivalent to array.dims
    df: None, number, or dict of {dim: d}
        spacing between elements of array (in frequency-space).
        None --> infer from ds if provided, else infer from array.
        number --> use this as df for all dims.
    rad: None or bool
        if True, interpret frequency-spacing (df) like it is "in radians",
            dividing it by 2 * pi before converting to position-space.
        None --> infer rad from names of the dims being ifft'd.
    ds: None, number, or dict of {dim: d}
        spacing between elements of result (in position-space), along dims from result.
        number --> use this as ds for all dims.
        None --> infer from df if provided, else infer from array.
        Note: provide ds to guarantee ifft(fft(arr)) == arr, exactly;
            otherwise position coords might include small rounding errors.
    x0: None, number, or dict of {dim: value}
        if provided, alter position-space coordinates by adding a constant offset,
            such that the 0'th position for each dim equals x0[dim].
        number --> apply the same number to all dims.
        iterable --> use these numbers; kwarg `dim` must also be provided as an iterable of strs.
        dict --> dict of {dim: x0} specifying the value associated with each dim
    '''
    # # parse input # #
    _dim_input = dim  # helpful for debugging; dim may be altered below.
    _rad_input = rad  # helpful for debugging. rad may be altered below.
    _df_input = df  # helpful for debugging. df may be altered below.
    _ds_input = ds  # helpful for debugging. ds may be altered below.
    # dim
    dims = array.dims if (dim is None) else ([dim] if isinstance(dim, str) else list(dim))
    # pos_dims
    fnames = []
    for dim in dims:
        fname = FFTDimname.implied_from(dim, array.coords, rad=rad, post_fft=True)  # (will crash if bad dim or rad)
        fnames.append(fname)
        rad = fname.rad  # <-- all fnames must have compatible rad!
    pos_dims = {fd.post: fd.pre for fd in fnames}  # {post-fft: pre-fft}. i.e., {pre-ifft: post-ifft}
    dims = list(pos_dims)  # dims is now the post-fft names. i.e., pre-ifft names.
    # promote non-dimension coordinates to dimensions, if necessary
    for dim in dims:
        array = xarray_promote_dim(array, dim)
    # df & ds
    if (df is not None) and (ds is not None):
        raise InputConflictError('provided df and ds. Expect ds, df, or neither, but not both!')
    if (df is None) and (ds is None):  # infer df from array coords
            df = {dim: xarray_get_dx_along(array, dim, float_rounding=True) for dim in dims}
    elif isinstance(df, str):  # raise helpful error message. Maybe caused by typo when inputting args.
        raise TypeError(f'expected None or dict-like df; got df={df!r}')
    if (df is not None) and (not is_iterable(df)):
        df = {dim: df for dim in dims}
    elif isinstance(df, dict):  # df is a dict, but user chose keys so they might not correspond to dims yet.
        df = {FFTDimname.implied_from(k, array.coords, rad=rad, post_fft=True).post: v for k, v in df.items()}
    if (ds is not None) and (not is_iterable(ds)):  # ds is a number
        ds = {dim: ds for dim in dims}
    elif isinstance(ds, dict):  # ds is a dict, but user chose keys so they might not correspond to dims yet.
        ds = {FFTDimname.implied_from(k, array.coords, rad=rad, post_fft=True).post: v for k, v in ds.items()}
    # x0
    if (x0 is not None) and (not is_iterable(x0)):  # x0 is a number
        x0 = {dim: x0 for dim in dims}
    elif isinstance(x0, dict):  # x0 is a dict, but user chose keys so they might not correspond to dims yet.
        x0 = {FFTDimname.implied_from(k, array.coords, rad=rad, post_fft=True).post: v for k, v in x0.items()}
    # # convert input to numpy-acceptable format # #
    np_array = array.values
    axes = tuple(array.dims.index(dim) for dim in dims)
    if df is not None: df = tuple(df.get(dim, None) for dim in dims)
    if ds is not None: ds = tuple(ds.get(dim, None) for dim in dims)
    if x0 is not None: x0 = tuple(x0.get(dim, None) for dim in dims)
    # # take the ifft (and handle the ifftshift, too) # #
    vals, pos = ifftN(np_array, df=df, axes=axes, rad=rad, x0=x0, ds=ds, **kw_np_ifftn)
    # # converting to xarray # #
    # dims (order matters, must match dims order in vals)
    result_dims = list(array.dims)
    for i, dim in enumerate(result_dims):
        if dim in pos_dims:
            result_dims[i] = pos_dims[dim]
    # coords (update any coordinate info relevant to ifft)
    pos_coords = {d: p for d, p in zip(dims, pos)}    # dict of {original freq dim name: associated pos values}
    result_coords = array.coords.copy(deep=False)
    for cname, cval in tuple(result_coords.items()):
        if cname in pos_dims:  # a dimension where we applied ifft.
            del result_coords[cname]
            pd = pos_dims[cname]
            result_coords[pd] = pos_coords[cname]
        elif any(cd in dims for cd in cval.dims):  # otherwise associated with a dimension where we applied ifft.
            del result_coords[cname]  # discard info -- ifft'd dims coords don't really map 1-to-1 to input coords.
        else:  # not associated with a dimension where we applied ifft.
            pass  # keep the coordinate information.
    attrs = array.attrs.copy() if xr.get_options()['keep_attrs'] else None
    result = xr.DataArray(vals, dims=result_dims, coords=result_coords, attrs=attrs)
    return result


''' --------------------- lowpass filtering --------------------- '''

@pcAccessor.register('lowpass', totype='array')
@format_docstring(default_lowpass=DEFAULTS.LOWPASS_KEEP)
def xarray_lowpass(array, dim=None, keep=UNSET, *, keep_r=UNSET, ds=None, real=None, return_fft=False):
    '''return array after putting it through a lowpass filter using fft & ifft.
    This is equivalent to ifft(fft(array) * filter), where filter is 0 at all "large" frequency values.
    "large" is determined by keep & r; see below.

    dim: None or iterable or strs
        coordinates to apply lowpass filter over. If None, use all array.dims.
        promote_dim(array, coord) for any non-dimension coordinates, as needed.
    keep: UNSET, True, dict, or number between 0 < keep <= 1
        fraction of frequencies to keep, along each dim.
        (Must provide this or keep_r but not both.)
        True --> use DEFAULTS.FFT_LOWPASS_KEEP.
        number --> use this value for all dims.
    keep_r: UNSET, True, or number between 0 < keep <= 1
        radius of N-sphere to keep in normalized frequency-space,
            normalized such that max(frequencies)==1 along each dim.
        All values outside of this N-sphere will be set to 0.
        Similar to `keep`, but here use an N-sphere instead of an N-cube.
        (Must provide this or keep but not both.)
        True --> use DEFAULTS.FFT_LOWPASS_KEEP.
        [TODO] more options than just spherical? (e.g. ellipsoid)
    ds: None, number, or dict of {{dim: d}}
        spacing between elements of array along each dim.
        number --> use the same value for all dims.
        None --> infer via array.coords[dim].diff(dim) for each dim
            (requires evenly-spaced coordinates in dim; spacing checked with np.allclose)
    real: None or bool
        whether to return np.real(ifft) instead of just ifft (which might have imaginary part)
        None --> infer from array. Use True if np.all(np.isreal(array)), else False.
    return_fft: bool
        whether to return (result, masked fft) instead of just result.
        mainly intended for debugging purposes.
    '''
    # # parse input # #
    _input_dim = dim   # remembering original inputs helps with debugging.
    _input_keep = keep
    _input_ds = ds
    # dim
    dims = array.dims if dim is None else dim
    for dim in dims:
        array = xarray_promote_dim(array, dim)
    # real
    if real is None:
        real = np.all(np.isreal(array))
    # ds
    if ds is None:
        ds = {dim: xarray_get_dx_along(array, dim, float_rounding=True) for dim in dims}
    # keep
    if keep is UNSET and keep_r is UNSET:
        raise InputMissingError('must provide keep or keep_r.')
    if keep is not UNSET and keep_r is not UNSET:
        raise InputConflictError('Expect keep or keep_r, but not both.')
    if keep is not UNSET:
        if keep is True:
            keep = DEFAULTS.LOWPASS_KEEP
        if not isinstance(keep, dict):
            keep = {dim: keep for dim in dims}
        post_keep = {FFTDimname(k).post: v for k, v in keep.items()}  # {post-fft dimname: keep}
    if keep_r is not UNSET:
        if keep_r is True:
            keep_r = DEFAULTS.LOWPASS_KEEP
    # # fft # #
    fft = xarray_fftN(array, dim=dims, ds=ds)
    fft_dims = [FFTDimname(d).post for d in dims]  # post-fft name for each dim in dims.
    # # lowpass filter # #
    # assumes that coords from xarray_fftN are centered around 0, and evenly spaced.
    if keep is not UNSET:  # set values outside of N-cube to 0.
        for dim, k in post_keep.items():
            # fft = 0 wherever |coords| > keep * max(|coords|)
            abs_coords = np.abs(fft.coords[dim])
            max_coord = np.max(abs_coords)
            fft = fft.where(abs_coords <= k * max_coord, other=0)  # values outside cube become 0.
    else:  # keep_r is not UNSET
        # fft = 0 wherever (distance from origin in normalized frequency space) >= keep_r
        dists = np.sqrt(sum((fft.coords[dim] / np.max(fft.coords[dim]))**2 for dim in fft_dims))
        fft = fft.where(dists <= keep_r, other=0)  # values outside sphere become 0.
    # # ifft # #
    result = xarray_ifftN(fft, dim=fft_dims, ds=ds)
    # # postprocessing # #
    if real:
        result = np.real(result)
    if return_fft:
        return result, fft
    return result



"""
File Purpose: testing tools from tools.xarray_tools.xarray_grids
"""
import pytest
import numpy as np
import xarray as xr

import PlasmaCalcs as pc


def test_xrrange():
    assert pc.xarray_range is pc.xrrange  # these are aliases for each other.
    xx = xr.DataArray(np.arange(4), coords={'x': np.arange(4)}, name='x')
    assert xx.identical(pc.xrrange(4, 'x'))
    assert xx.identical(pc.xrrange(4, name='x'))  # can supply name as kwarg instead of positional
    xx = xr.DataArray(np.arange(1, 5), coords={'x': np.arange(1, 5)}, name='x')
    assert xx.identical(pc.xrrange((1, 5), 'x'))
    xx = xr.DataArray(np.arange(1, 9, 2), coords={'x': np.arange(1, 9, 2)}, name='x')
    assert xx.identical(pc.xrrange((1, 9, 2), 'x'))
    with pytest.raises(pc.InputError):
        pc.xrrange(1, 5)  # name must be str. E.g., user forgot to put (1,5) as tuple.
    # ensure can create array even without name; uses a default name.
    xx = pc.xrrange(5)
    # ensure can provide coords
    xx = pc.xrrange(5, 'x', coords={'y': 7})
    assert xx.coords['y'] == 7
    assert np.all(xx.coords['x'] == np.arange(5))
    # ensure coords can include name
    xx = pc.xrrange(5, 'x', coords={'x': 10+np.arange(5), 'y': 7})
    assert xx.coords['y'] == 7
    assert np.all(xx.coords['x'] == 10+np.arange(5))

def test_xarray_grid():
    assert np.all(pc.XarrayGrid(1, 100, logstep=1)() == [1,10,100])
    assert np.all(pc.XarrayGrid(1, 10, step=3).grid() == [1,4,7,10])  # grid() is alias for calling
    with pytest.raises(pc.InputConflictError):
        pc.XarrayGrid(pc.xrrange(5, 'x'), 10, step=2)  # can't provide both step and min & max.
    arr = pc.XarrayGrid(pc.xrrange(3, 'x'), span=6, step=2, name='mygrid')()
    assert np.all(arr.isel(x=0) == [0, 2, 4, 6])
    assert np.all(arr.isel(x=1) == [1, 3, 5, 7])
    assert np.all(arr.isel(x=2) == [2, 4, 6, 8])
    assert arr.name == 'mygrid'
    assert set(arr.dims) == set(('x', 'mygrid_dim'))
    assert 'mygrid' in arr.coords
    assert arr['mygrid'].dims == arr.dims
    assert pc.XarrayGrid(1, 100, 50)().identical(pc.xarray_grid(1, 100, 50))
    assert pc.XarrayGrid(1, 100, 50, logspace=True)().identical(pc.xarray_grid(1, 100, 50, logspace=True))
    # step not matching span (or ratio, if logspace) exactly:
    with pytest.raises(pc.InputConflictError):
        pc.xarray_grid(0, 1, step=0.3)  # <-- step doesn't divide evenly into span.
    assert np.allclose(pc.xarray_grid(0, 1, step=0.3, inclusive=(True, False)), [0, 0.3, 0.6, 0.9])
    assert np.allclose(pc.xarray_grid(0, 1, step=0.3, inclusive=(False, True)), [0.1, 0.4, 0.7, 1])
    assert np.allclose(pc.xarray_grid(0, 1, step=0.3, inclusive=(False, False)), [0.3, 0.6, 0.9])
    assert np.all(pc.xarray_grid(10, 17, step=2, inclusive=(True, False)) == [10, 12, 14, 16])
    assert np.all(pc.xarray_grid(10, 17, step=2, inclusive=(False, True)) == [11, 13, 15, 17])
    assert np.all(pc.xarray_grid(10, 17, step=2, inclusive=(False, False)) == [12, 14, 16])
    with pytest.raises(pc.InputConflictError):
        pc.xarray_grid(1, 900, logstep=0.5)  # <-- logstep doesn't divide evenly into span.
    loggrid = np.logspace(0, 3, 7)  # [1, 3.16, 10, 31.6, 100, 316, 1000]
    assert np.allclose(pc.xarray_grid(1, 900, logstep=0.5, inclusive=(True, False)), loggrid[:-1])
    assert np.allclose(pc.xarray_grid(1, 900, logstep=0.5, inclusive=(False, False)), loggrid[1:-1])
    assert np.allclose(pc.xarray_grid(1, 900, logstep=1, inclusive=(True, False)), [1, 10, 100])
    assert np.allclose(pc.xarray_grid(1, 900, logstep=1, inclusive=(False, False)), [10, 100])
    assert np.allclose(pc.xarray_grid(1, 900, logstep=1, inclusive=(False, True)), [9, 90, 900])
    # grid smaller than allowed:
    with pytest.raises(pc.DimensionSizeError):
        pc.xarray_grid(0, 100, N=1)
    with pytest.raises(pc.DimensionSizeError):
        pc.xarray_grid(0, 100, N=0)
    assert np.all(pc.xarray_grid(0, 100, step=100)==[0,100])
    with pytest.raises(pc.DimensionSizeError):
        pc.xarray_grid(0, 100, step=100, inclusive=(True, False))
    with pytest.raises(pc.DimensionSizeError):
        pc.xarray_grid(0, 100, step=100, inclusive=(False, False))
    with pytest.raises(pc.DimensionSizeError):
        pc.xarray_grid(0, 100, step=50, inclusive=(False, False))
    # grid larger than allowed:
    with pytest.raises(pc.DimensionSizeError):  # <-- ensure the N_max limitation works on a small test case
        pc.xarray_grid(0, 1, step=1e-6, N_max=1000)   # (artificially lower N_max)
    with pytest.raises(pc.DimensionSizeError):  # <-- ensure the N_max limitation works on a huge case
        pc.xarray_grid(0, 1, step=1e-10)  # this grid would take up ~80 GB, if N_max didn't cause crash first!



"""
File Purpose: basic tests for eppic hookup in PlasmaCalcs.
These tests require some simulation output.
They also test reading lots of different basic values,
    with lots of different values for dims & slices,
    so it takes a little while to run (~20-30 seconds?).
"""
import os
import pytest

import numpy as np

import PlasmaCalcs as pc
pc.DEFAULTS.pic_ambiguous_unit = dict(u_t=1)  # must be defined before loading an EppicCalculator.

HERE = os.path.dirname(__file__)


''' --------------------- can create eppic calculator --------------------- '''

def get_eppic_calculator(**kw_init):
    '''make an eppic calculator'''
    with pc.InDir(os.path.join(HERE, 'test_eppic_tinybox')):
        ec = pc.EppicCalculator.from_here(**kw_init)
    return ec


''' --------------------- basic tests --------------------- '''

def test_dimensions_manip():
    '''ensure can manipulate dimensions properly.'''
    ec = get_eppic_calculator()
    # snaps: there are 10 snaps: 0,1,...,9. Going from it='2560' to it='25600'
    ec.snap = None; assert ec.snap.size == 10
    ec.snap = 0; assert ec.snap.size == 1
    ec.snap = '2560'; assert ec.snap.size == 1
    ec.snap = ['5120', '7680']; assert len(ec.snap) == 2
    ec.snap = 4; assert ec.snap.size == 1
    ec.snap = slice(0, None, 2); assert ec.snap.size == 5
    ec.snap = slice(0, None, 0.4)  # check that "interprets_fractional_indexing" works as expected
    assert ec.snap == ec.snaps[slice(0, None, 4)]
    assert ec.snap.size == 3
    ec.snap = -1; assert ec.snap.size == 1
    # fluids: there are 3 fluids ('e-', 'H+', 'C+')
    ec.fluid = None; assert ec.fluid.size == 3; assert ec.fluid == ['e-', 'H+', 'C+']
    ec.fluid = 0; assert ec.fluid.size == 1; assert ec.fluid == 'e-'
    ec.fluid = -1; assert ec.fluid.size == 1; assert ec.fluid == 'C+'
    ec.fluid = 'H+'; assert ec.fluid.size == 1
    ec.fluid = ['H+', 'C+']; assert len(ec.fluid) == 2
    ec.fluid = ec.fluids.get_electron(); assert ec.fluid.size == 1
    ec.fluid = ec.fluids.charged(); assert ec.fluid.size == 3
    with pytest.raises(pc.FluidKeyError):  # eppic has no neutral fluid, only a constant background.
        ec.fluids.get_neutral()
    # components: there are 3 components ('x', 'y', 'z')
    ec.component = None; assert ec.component.size == 3; assert ec.component == ['x', 'y', 'z']
    ec.component = 0; assert ec.component.size == 1; assert ec.component == 'x'
    ec.component = 'z'; assert ec.component.size == 1
    ec.component = ['x', 'y']; assert len(ec.component) == 2
    ec.component = slice(0, None, 2); assert ec.component.size == 2
    # maindims: there are 2 maindims ('x', 'y')  (it was a 2D run) each with size 16.
    assert len(ec.maindims) == 2
    ec.slices = dict(); assert ec.maindims_shape == (16, 16); assert ec.maindims_size == 16*16
    ec.slices = dict(x=0); assert ec.maindims_shape == (1, 16); assert ec.maindims_size == 16
    ec.slices = dict(x=5); assert ec.maindims_shape == (1, 16); assert ec.maindims_size == 16
    ec.slices = dict(y=-1); assert ec.maindims_shape == (16, 1); assert ec.maindims_size == 16
    ec.slices = dict(x=0, y=0); assert ec.maindims_shape == (1, 1); assert ec.maindims_size == 1
    ec.slices = dict(x=slice(0, None, 2)); assert ec.maindims_shape == (8, 16); assert ec.maindims_size == 8*16
    ec.slices = dict(x=slice(0, 10, 2), y=slice(7, 11)); assert ec.maindims_shape == (5, 4); assert ec.maindims_size == 5*4
    # check that "interprets_fractional_indexing" works as expected
    ec.slices = dict(x=slice(0, 0.5)); assert ec.maindims_shape == (8, 16)
    ec.slices = dict(x=slice(-0.5, None)); assert ec.maindims_shape == (7, 16)
    ec.slices = dict(x=slice(None, None, 0.25), y=slice(None, None, 0.125)); assert ec.maindims_shape == (4, 8)

# takes roughly 20 seconds
def test_reading_basic_values():
    '''ensure can read basic values from eppic.'''
    # checks that output size is correct, for various dimensions.
    # use some of the same dimensions as in test_dimensions_manip, to enable hard-coding the expected results.
    # don't use all of them, because that produces hundreds of input combinations, which takes a while.
    # also compare against the dynamically-calculated expected results.
    ec = get_eppic_calculator()
    snap_inputs = [None, 4, slice(0, None, 2)]
    snap_sizes = [10, 1, 5]
    fluid_inputs = [None, -1, ['H+', 'C+']]
    fluid_sizes = [3, 1, 2]
    component_inputs = [None, 'z', ['x', 'y']]
    component_sizes = [3, 1, 2]
    slice_inputs = [dict(), dict(x=5), dict(x=0, y=0), dict(x=slice(0, 10, 2), y=slice(7, 11))]
    slice_shapes = [(16, 16), (1, 16), (1, 1), (5, 4)]
    slice_sizes = [16*16, 16, 1, 5*4]

    input_attrs = []  # list of dicts of attr values
    input_sizes = []  # list of dicts of attr sizes
    for snap, snap_size in zip(snap_inputs, snap_sizes):
        for fluid, fluid_size in zip(fluid_inputs, fluid_sizes):
            for component, component_size in zip(component_inputs, component_sizes):
                for slices, slice_size in zip(slice_inputs, slice_sizes):
                    input_attrs.append(dict(snap=snap, fluid=fluid, component=component, slices=slices))
                    input_sizes.append(dict(snap=snap_size, fluid=fluid_size, component=component_size, slice=slice_size))
    assert len(input_attrs) == len(input_sizes)
    assert len(input_attrs) == len(snap_sizes) * len(fluid_sizes) * len(component_sizes) * len(slice_sizes)

    for attrs, sizes in zip(input_attrs, input_sizes):
        ec.set_attrs(**attrs)
        # redundant with test_dimensions_manip to check size here, but these are tests so redundancy is fine.
        assert ec.current_n_snap() == sizes['snap']
        assert ec.current_n_fluid()== sizes['fluid']
        assert ec.current_n_component() == sizes['component']
        assert ec.maindims_size == sizes['slice']
        # get some basic values and assert they have the correct size.
        arr = ec('n')  # in its own line for easier debugging in case of crash.
        dynamic_expect_size = ec.match_var_result_size('n')  # dynamically calculate expected size.
        simple_expect_size = sizes['snap'] * sizes['fluid'] * sizes['slice']  # size based on the things 'n' depends on.
        assert arr.size == dynamic_expect_size == simple_expect_size
        arr = ec('E')
        dynamic_expect_size = ec.match_var_result_size('E')
        simple_expect_size = sizes['snap'] * sizes['component'] * sizes['slice']
        assert arr.size == dynamic_expect_size == simple_expect_size
        arr = ec('u')
        dynamic_expect_size = ec.match_var_result_size('u')
        simple_expect_size = sizes['snap'] * sizes['fluid'] * sizes['component'] * sizes['slice']
        assert arr.size == dynamic_expect_size == simple_expect_size
        arr = ec('Ta')
        dynamic_expect_size = ec.match_var_result_size('Ta')
        simple_expect_size = sizes['snap'] * sizes['fluid'] * sizes['component'] * sizes['slice']
        assert arr.size == dynamic_expect_size == simple_expect_size
        arr = ec('m')
        dynamic_expect_size = ec.match_var_result_size('m', maindims=False)  # [TODO] infer maindims=False from 'm'
        simple_expect_size = sizes['fluid']
        assert arr.size == dynamic_expect_size == simple_expect_size
        arr = ec('m_neutral')
        dynamic_expect_size = ec.match_var_result_size('m_neutral', maindims=False) # [TODO] infer maindims=False from 'm'
        simple_expect_size = 1
        assert arr.size == dynamic_expect_size == simple_expect_size
        # get some slightly more complicated values and assert they have the correct size.
        arr = ec('E_x')
        dynamic_expect_size = ec.match_var_result_size('E_x')
        simple_expect_size = sizes['snap'] * sizes['slice']
        assert arr.size == dynamic_expect_size == simple_expect_size
        arr = ec('u_y')
        dynamic_expect_size = ec.match_var_result_size('u_y')
        simple_expect_size = sizes['snap'] * sizes['fluid'] * sizes['slice']
        assert arr.size == dynamic_expect_size == simple_expect_size
        arr = ec('T')
        dynamic_expect_size = ec.match_var_result_size('T')
        simple_expect_size = sizes['snap'] * sizes['fluid'] * sizes['slice']
        assert arr.size == dynamic_expect_size == simple_expect_size
        arr = ec('mod_E')
        dynamic_expect_size = ec.match_var_result_size('mod_E')
        simple_expect_size = sizes['snap'] * sizes['slice']
        assert arr.size == dynamic_expect_size == simple_expect_size
        arr = ec('mean_n')
        dynamic_expect_size = ec.match_var_result_size('mean_n', maindims=False)  # [TODO] infer maindims=False from 'mean'.
        simple_expect_size = sizes['snap'] * sizes['fluid']
        assert arr.size == dynamic_expect_size == simple_expect_size
        arr = ec('std_u')
        dynamic_expect_size = ec.match_var_result_size('std_u', maindims=False)  # [TODO] infer maindims=False from 'std'.
        simple_expect_size = sizes['snap'] * sizes['fluid'] * sizes['component']
        assert arr.size == dynamic_expect_size == simple_expect_size

def test_reading_fractional_slices():
    '''test reading values when using fractional slices'''
    ec = get_eppic_calculator()
    ec.snap = 7  # choosing a single snap is sufficient for these tests.
    ec.fluid = 0  # choosing a single fluid is sufficient for these tests.
    assert ec.maindims_shape == (16, 16)
    # hard-coded translations below, between fractional & integer slices, assumes maindims_shape==(16,16)
    assert ec('n', slices=dict(x=slice(0, 0.5))).identical(ec('n').isel(x=slice(0, 8)))
    assert ec('n', slices=dict(x=slice(-0.5, None))).identical(ec('n').isel(x=slice(-7, None)))
    assert ec('n', slices=dict(y=slice(0.25, 0.75))).identical(ec('n').isel(y=slice(4, 12)))
    assert ec('n', slices=dict(y=slice(-0.25, None))).identical(ec('n').isel(y=slice(-3, None)))
    assert ec('n', slices=dict(x=slice(None, None, 0.25))).identical(ec('n').isel(x=slice(None, None, 4)))
    # also check that it is possible to use negative step:
    with pytest.raises(NotImplementedError):  # negative step not yet implemented.
        ec('n', slices=dict(x=slice(None, None, -1)))
    # assert ec('n', slices=dict(x=slice(None, None, -1))).identical(ec('n').isel(x=slice(None, None, -1)))
    # assert ec('n', slices=dict(x=slice(None, None, -0.25))).identical(ec('n').isel(x=slice(None, None, -4)))
    # for slice_ in [slice(10, 5, -1), slice(16, 3, -2), slice(0, 5, -1), slice(7, None, -0.25)]:
    #     assert ec('n', slices=dict(x=slice_)).identical(ec('n').isel(x=slice_))

def test_vector_arithmetic():
    '''some tests of vector arithmetic methods'''
    ec = get_eppic_calculator()
    ec.snap = 5  # choosing a single snap is sufficient for these tests.
    # ensure perpmod actually independent of component
    E_perpmod_B = ec('E_perpmod_B')
    assert E_perpmod_B.identical(ec('E_perpmod_B', component='x'))
    assert E_perpmod_B.identical(ec('E_perpmod_B', component=['x', 'y']))
    # ensure perpmod nonzero (not necessarily true in general, but true for this data) & non-nan.
    assert np.all(E_perpmod_B > 0)
    assert np.all(np.isfinite(E_perpmod_B))
    # ensure perpmod identical to perpmag (they are aliases)
    assert E_perpmod_B.equals(ec('E_perpmag_B'))  # can't use "identical" because names are different.
    # ensure for perp, component affects result in the expected way
    E_perp_B = ec('E_perp_B')
    assert E_perp_B.isel(component=0).identical(ec('E_perp_B', component='x'))
    assert np.all(ec('(E_perp_B)_x') > 0)  # not generally true for all data, but true for this data
    assert np.all(ec('E_perp_B', component='y') < 0)  # similarly, should be true for this particular data

def test_deriv_before_slice():
    '''test that deriv_before_slice works as expected.'''
    ec = get_eppic_calculator()
    ec.snap = 3  # choosing a single snap is sufficient for these tests.
    ec.fluid = 0  # choosing a single fluid is sufficient for these tests.
    SLICES_TO_TEST = [
        dict(),
        dict(x=5),
        dict(x=slice(3, 15, 1)),
        dict(x=slice(0, 10, 2)),
        dict(x=slice(0, 10, 2), y=slice(7, 11)),
        dict(x=slice(0, 0.8, 0.25)),
    ]
    for var in ('ddx_n', 'grad_n', 'div_u', 'curl_u'):
        for slices in SLICES_TO_TEST:
            simple_before = ec(var, slices=dict()).pc.isel(slices)  # like deriv_before_slice=True
            deriv_before = ec(var, deriv_before_slice=True, slices=slices)
            deriv_after = ec(var, deriv_before_slice=False, slices=slices)
            assert simple_before.identical(deriv_before)
            if slices:  # i.e. slices != empty dict
                assert not np.allclose(deriv_after, deriv_before)
            else:  # i.e. no slices applied
                assert deriv_after.identical(deriv_before)
        # test mode='step', when step!=1:
        slices = dict(x=slice(5, 13, 2))  # could be any slice with step!=1.
        simple_before = ec(var, slices=dict()).pc.isel(slices)  # like deriv_before_slice=True
        step_after = ec(var, slices=dict(x=slice(5, 13, None))).pc.isel(x=slice(None, None, 2))
        deriv_after = ec(var, deriv_before_slice=False, slices=slices)
        deriv_step = ec(var, deriv_before_slice='step', slices=slices)
        deriv_before = ec(var, deriv_before_slice=True, slices=slices)
        assert simple_before.identical(deriv_before)
        assert not np.allclose(deriv_after, deriv_before)
        assert step_after.identical(deriv_step)
        assert not np.allclose(deriv_step, deriv_before)
        assert not np.allclose(deriv_step, deriv_after)
        # test mode='step', when step==1 (or None):
        slices = dict(x=slice(3, 8, 1))  # could be any slice with step==1.
        simple_before = ec(var, slices=dict()).pc.isel(slices)
        step_after = ec(var, slices=dict(x=slice(3, 8, None))).pc.isel(x=slice(None, None, 1))
        deriv_after = ec(var, deriv_before_slice=False, slices=slices)
        deriv_step = ec(var, deriv_before_slice='step', slices=slices)
        deriv_before = ec(var, deriv_before_slice=True, slices=slices)
        assert simple_before.identical(deriv_before)
        assert not np.allclose(deriv_after, deriv_before)
        assert step_after.identical(deriv_step)
        assert not np.allclose(deriv_step, deriv_before)
        assert deriv_step.identical(deriv_after)  # deriv_step == deriv_after when step==1.
        # test mode='step' when step is fractional
        slices = dict(x=slice(0, 13, 0.25))
        assert ec.maindims_shape == (16, 16)
        step_after = ec(var, slices=dict(x=slice(0, 13, None))).pc.isel(x=slice(None, None, int(0.25*16)))
        deriv_step = ec(var, deriv_before_slice='step', slices=slices)
        assert step_after.identical(deriv_step)
        # test mode='step' when using int or list indexer instead of slice
        xslices = [dict(x=5), dict(x=[0, 7, 10])]
        for slices in xslices:
            assert ec.maindims_shape == (16, 16)
            deriv_step = ec(var, deriv_before_slice='step', slices=slices)
            deriv_before = ec(var, deriv_before_slice=True, slices=slices)
            assert deriv_step.identical(deriv_before)



"""
File Purpose: defaults in PlasmaCalcs
"""
import numpy as np

class _Defaults():
    '''stores defaults for PlasmaCalcs. Use DEFAULTS instead.
    (DEFAULTS is an instance of _Defaults(), instantiated at the bottom of defaults.py)
    '''
    def update(self, other_defaults):
        '''update self with other_defaults, overwriting any existing values.
        other_defaults: dict or _Defaults instance
            dict --> update from other_defaults.items()
            else --> update from other_defaults.__dict__.items()
        '''
        if not isinstance(other_defaults, dict):
            other_defaults = other_defaults.__dict__
        for key, value in other_defaults.items():
            setattr(self, key, value)

    TRACEBACKHIDE = True
    IMPORT_FAILURE_WARNINGS = False

    TAB = ' '*4   # size of a single tab, e.g. for docstrings.

    DEBUG = 0   # level of debugging; 0 = no debugging.
    PROGRESS_UPDATES_PRINT_FREQ = 2  # seconds between progress updates

    # skip these attrs during Behavior.label_array(), by default.
    SKIP_BEHAVIOR_ARRAY_LABELS = []

    # size limit for CachesLoader results writable to _pc_caches file.
    # crash with MemorySizeError before writing array larger than this to cache file.
    # (to set to no limit, use None)
    # default 1 MB easily allows <100,000 floats, which is great for 1D arrays,
    #   or timelines, e.g. mean_u across multiple fluids & components,
    #   but not good for values across maindims.
    CACHE_ARRAY_MBYTES_MAX = 1  # [MB]

    # raise MemorySizeError before loading arrays larger than this, by default.
    # (to set to no limit, use None)
    ARRAY_MBYTES_MAX = 1000  # [MB]

    # during memory_size_check, pretend array alements are at least as memory-intensive as this dtype:
    # (only used if array elements are smaller than this dtype.) (to set to no minimum, use None)
    ARRAY_MEMORY_CHECK_DTYPE_MIN = np.float64

    # during load_across_dims, use this many cpus by default, if ncpu not provided.
    # use None to automatically determine limit (based on number of cores available)
    LOADING_NCPU = 1
    # during load_across_dims, timeout after this many seconds, by default, if timeout not provided.
    # timeout numbers must be an integer (due to limitations of signal.alarm method).
    # For no time limit, use None, 0, or negative number.
    LOADING_TIMEOUT = None
    # during load_across_dims, if ncpu>1, group tasks into groups of size ncoarse before performing them.
    # use ncoarse=1 to avoid any coarsening.
    LOADING_NCOARSE = 1
    # during load_across_dims, use this builtins.multiprocessing.pool.Pool instead of making a new one.
    # (if provided, ncpu will be ignored.)
    # use None to make a new pool each time, when needed.
    LOADING_POOL = None

    # [EFF] when stats_dimpoint_wise is None, this tells min size before using stats_dimpoint_wise.
    #    if result would have size less than this, first try stats_dimpoint_wise=False.
    #    (this includes maindims shape.)
    #    (stats_dimpoint_wise True seems to be faster for larger arrays but slower for smaller arrays.)
    #    None --> no minimum, i.e. always prefer stats_dimpoint_wise = False.
    STATS_DIMPOINT_WISE_MIN_N = 512 * 512 * 200
    # [EFF] when stats_dimpoint_wise, this tells min length of dimension before loading across it,
    #   in the first attempt to load across dims implied. 1 --> no minimum.
    #   (if the first attempt fails, will repeat but with min_split=1.)
    STATS_DIMPOINT_WISE_MIN_SPLIT = 10

    # for displaying trees. See Tree._html for details.
    TREE_CSS = '''
    <style type="text/css">
    summary {display: list-item}          /* show arrows for details/summary tags */
    details > *:not(summary){margin-left: 2em;}   /* indent details/summary tags. */
    </style>
    '''
    TREE_SHOW_DEPTH = 4  # default depth to show when displaying trees.
    TREE_SHOW_MAX_DEPTH = 50  # max depth to render as html when displaying trees.
    TREE_SHORTHAND = False  # whether to use shorthand for Tree nodes when displaying trees.

    # fft renaming convention for dimensions, when rad=True, and freq_dims not provided.
    # dict of {old dim: new dim} pairs. Or, None, equivalent to empty dict.
    # for dims not appearing here, new dim will be 'freqrad_{dim}'.
    FFT_FREQ_RAD_DIMNAMES = {    
        'x': 'k_x',
        'y': 'k_y',
        'z': 'k_z',
        't': 'omega',
    }
    # default value for "keep" in xarray_fftN. Should be 0 < value <= 1.
    FFT_KEEP = 0.4
    # default value for "keep" in xarray_lowpass. Should be 0 < value <= 1.
    LOWPASS_KEEP = 0.4

    # default method to use for rounding fractional indexes to integers, as per interprets_fractional_indexing.
    # 'round' --> as per builtins.round(). round to nearest integer, ties toward even integers.
    # 'int' --> as per builtins.int(). round towards 0.
    # 'floor' --> as per math.floor(). round towards negative infinity.
    # 'ceil' --> as per math.ceil(). round towards positive infinity.
    # note that 'floor' is the default to avoid accidentally accessing using index=len(array).
    FRACTIONAL_INDEX_ROUNDING = 'floor'

    # default for gaussian filter sigma, if not provided.
    GAUSSIAN_FILTER_SIGMA = 1.0

    # if this number density in [m^-3] is too large for float32 in output units system,
    # then convert to float64. Commonly relevant if calculating 'n' in 'raw' units for an MhdCalculator.
    MHD_MAX_SAFE_N_SI = 1e30


class _PlotDefaults():
    '''stores defaults for PlasmaCalcs. Use DEFAULTS.PLOT instead.
    (DEFAULTS.PLOT is an instance of _PlotDefaults(), instantiated at the bottom of defaults.py)
    '''
    ASPECT = 'equal'  # default aspect ratio for plots
    LAYOUT = 'compressed'   # default layout for plots. 'constrained', 'compressed', 'tight', or 'none'

    # default for robust when determining vmin & vmax. True, False, or number between 0 and 50.
    # True --> use DEFAULTS.PLOT.ROBUST_PERCENTILE.
    ROBUST = True
    # default percentile for determining vmin & vmax, when using robust.
    ROBUST_PERCENTILE = 2.0

    # default for min_n_ticks on x and y axes
    # None --> use matplotlib default (which is 2, as of 2025-03-20).
    # tuple of (nx, ny) --> use nx as min for x-axis, ny as min for y-axis.
    MIN_N_TICKS = 3
    # whether it's okay for set_min_n_ticks() to fail silently within PlasmaCalcs plotting routines.
    MIN_N_TICKS_FAIL_OK = False

    # default for min_n_ticks on colorbars.
    # None --> use matplotlib default (which is 2, as of 2025-03-20).
    # tuple of (nx, ny) --> use nx as min for horizontal colorbars, ny as min for vertical colorbars.
    MIN_N_TICKS_CBAR = 3
    # whether it's okay for set_min_n_ticks() to fail silently when applied to colorbars
    MIN_N_TICKS_CBAR_FAIL_OK = True

    # # params for plotting.colorbar.make_cax # #
    CAX_PAD = 0.01
    CAX_SIZE = 0.02
    CAX_LOCATION = 'right'
    # # other colorbar params # #
    CAX_MODE = 'mpl'   # 'mpl' or 'pc'. 'mpl' uses matplotlib logic to make cax; 'pc' uses PlasmaCalcs.make_cax.
                        # note that 'pc' looks better if using layout='none', but worse with any other layout.

    # # params for movies # #
    FPS = 30   # frames per second (except for small movies). if None, use matplotlib defaults.
    BLIT = True  # whether to use blitting. if None, use matplotlib defaults.
    MOVIE_EXT = '.mp4'  # movie filename extension to use if none provided. if None, use matplotlib defaults.
    FPS_SMALL = 2  # default frames per second for small movies. if None, use matplotlib defaults.
    NFRAMES_SMALL = 20  # movies with this many or fewer frames use FPS_SMALL by default.

    MOVIE_TITLE_FONT = 'monospace'  # font for movie titles

    # [seconds] minimum time between progress updates when saving movie.
    # use 0 for no minimum; use -1 for "never print".
    MOVIE_PROGRESS_UPDATE_FREQ = 1

    # whether to print help message about how to display movie inline,
    # if it might be applicable (e.g., using ipython) and plt.rcParams['animation.html']=='none'.
    MOVIE_REPR_INLINE_HELP = True

    # # params for subplots # #
    # [chars] suggested width of titles for subplots;
    # some routines might make multiline title if title would be longer than this.
    SUBPLOT_TITLE_WIDTH = 20  

    # [chars] suggested width of suptitle
    # some routines might make multiline suptitle if suptitle would be longer than this.
    SUPTITLE_WIDTH = 40

    # default kwargs for rtitles in subplots
    RTITLE_KW = {
        'rotation': 270,
        'loc': 'outside center right',
        'fontsize': 'large',
    }

    # default kwargs for ttitles in subplots
    TTITLE_KW = {}

    # # other params # #
    # DIMS_INFER tell which array.dims to use for x axis, y axis, and time axis (if movie).
    #   each list will be checked in order, using first match in case of multiple matches.
    #   This only applies when trying to infer plot dims which were not specified.
    DIMS_INFER = {
        # plot time axis
        't': ('time', 't', 'snap', 'snapshot', 'frame'),
        # plot x axis
        'x': ('x', 'r', 'y', 'z',
              'freq_x', 'freq_y', 'freq_z',
              'kx', 'ky', 'kz',
              'k_x', 'k_y', 'k_z',
              'kt', 'k_t', 'ktheta', 'k_theta', 'kang', 'k_ang',
              ),
        # plot y axis
        'y': ('y', 'z',
              'freq_y', 'freq_z',
              'k_y', 'k_z', 'k_mod', 'kmod', 'log_kmod',
              ),
    }
    # if dim in any set in DIMS_SAME, then the other dims are "redundant" with that dim.
    # usually not used, but e.g. in infer_subplot_title, if t_plot_dim is 'snap' or 't',
    # don't include the other one in the subplot title either, since they appear together in a set in DIMS_SAME.
    DIMS_SAME = [
        {'time', 'snap', 't'},
    ]

    SUBPLOTS_MAX_NROWS_NCOLS = 15  # max number of rows or cols in a subplots grid before crashing.
    SUBPLOTS_AXSIZE = (2,2)  # default size of each subplot, in inches.

    # # params for plotting.timelines # #
    # maximum length of a dimension before crashing when plotting timelines.
    # (helps to avoid accidentally creating one line for each x,y,z coords...)
    # e.g. if 10, when plotting 'fluid' and 'component', require len(fluids)<=10 and len(component)<=10.
    TIMELINES_DMAX = 10  # use None for no limit.

    # cycles for timelines. each cycle can be:
    #   dict --> interpretted as {matplotlib kwarg: list of values},
    #   None --> plt.rcParams['axes.prop_cycle']
    #   Cyler (from cycler module)
    TIMELINES_CYCLE0 = None  # default cycle for first dim. 
    TIMELINES_CYCLE1 = {'ls': ['-', '--', ':', '-.', (0,(3,1,1,1,1,1))]}  # default cycle for second dim.
    #  (0,(3,1,1,1,1,1)) means "densely dashdotdotted".

    # max length of strings for xticklabels. Longer than this will be cutoff with ...
    XTICKLABEL_MAXLEN = 15

    # # params for scatter plots # #
    # generic scatter style
    SCATTER_STYLE = {
        'linewidths': 2,
    }
    # style for scatter_max()
    SCATTER_MAX = {
        'marker': 'x',
        'color': 'red',
    }
    # style for scatter_min()
    SCATTER_MIN = {
        'marker': 'o',
        'facecolors': 'none',
        'color': 'red',
    }

    # # params for faceplot # #
    # default viewing angle for 3D faceplots, as (elevation, azimuth, roll).
    # (-160, 30, 0) provides a decent angle for viewing x=0, y=0, z=0 faces.
    FACEPLOT_VIEW_ANGLE = (-160, 30, 0)

    # default kwargs for ax.plot of edges in 3D faceplots.
    # set to None to not plot edges.
    FACEPLOT_EDGE_KWARGS = {'color': '0.4', 'linewidth': 1, 'zorder': 1e3}

    # faceplot axes zoom factor. Default 1. Must be >0. See ax.set_box_aspect for details.
    FACEPLOT_AXES_ZOOM = 1.0

    # aspect for 3D plots
    # 'auto', 'equal', (x aspect, y aspect, z aspect),
    # or (1, x multiplier, y multiplier, z multiplier);
    #    multiplier multiplies aspect determined by data lengths. >1 --> longer.
    ASPECT3D = 'equal'

    # projection type for 3D plots. 'ortho' or 'persp'
    # For more details see mpl_toolkits.mplot3d.axes3d.Axes3D
    PROJ_TYPE = 'ortho'

    # for contour plots, if using colorbar, linewidth of lines in colorbar.
    # None --> use same width as contour lines.
    # 2-tuple of None or int: defines (min, max); None for no bound.
    #   E.g. (4, None) says "for thinner lines use 4; others same as contour lines".
    COLORBAR_LINEWIDTH = (4, None)

    # for contour plots, if using colorbar, linestyle of lines in colorbar.
    # None --> use same linestyle as contour lines.
    COLORBAR_LINESTYLE = None


class _PhysicalDefaults():
    '''stores physical values defaults for PlasmaCalcs. Use DEFAULTS.PHYSICAL instead.
    (DEFAULTS.PHYSICAL is an instance of _PhysicalDefaults(), instantiated at the bottom of defaults.py)
    '''
    # values of various physical constants in SI units
    CONSTANTS_SI = {
        'amu'     : 1.66054e-27,  # atomic mass unit
        'c'       : 2.99792E8,    # speed of light
        'kB'      : 1.38065E-23,  # boltzmann constant
        'eps0'    : 8.85419E-12,  # permittivity of free space
        'mu0'     : 1.256637E-6,  # permeability of free space
        'qe'      : 1.60218E-19,  # elementary charge
        'me'      : 9.10938E-31,  # electron mass
        'qme'     : 1.75882E11,   # q / m_e
        'hplanck' : 6.260701E-34, # planck constant (not hbar)
        'm_proton': 1.67262E-27,  # proton mass
        'eV'      : 1.602176634E-19,    # electron volt
        # also, see below for some additional derived constants:
    }
    CONSTANTS_SI['eV kB-1']  = CONSTANTS_SI['eV'] / CONSTANTS_SI['kB']
    CONSTANTS_SI['me amu-1'] = CONSTANTS_SI['me'] / CONSTANTS_SI['amu']
    for _key, _alias in (('qe', 'q_e'), ('me', 'm_e'), ('m_proton', 'm_p'), ('hplanck', 'h')):
        CONSTANTS_SI[_alias] = CONSTANTS_SI[_key]
    del _key, _alias

    # atomic weight [amu] of each element
    M_AMU = {
        'H' :  1.008,
        'He':  4.003,
        'C' : 12.01,
        'N' : 14.01,
        'O' : 16.0,
        'Ne': 20.18,
        'Na': 23.0,
        'Mg': 24.32,
        'Al': 26.97,
        'Si': 28.06,
        'S' : 32.06,
        'K' : 39.1,
        'Ca': 40.08,
        'Cr': 52.01,
        'Fe': 55.85,
        'Ni': 58.69,
        }

    # first ionization potential [eV] of each element
    IONIZE_EV = {
        'H' : 13.595,
        'He': 24.580,
        'C' : 11.256,
        'N' : 14.529,
        'O' : 13.614,
        'Ne': 21.559,
        'Na':  5.138,
        'Mg':  7.644,
        'Al':  5.984,
        'Si':  8.149,
        'S' : 10.357,
        'K' :  4.339,
        'Ca':  6.111,
        'Cr':  6.763,
        'Fe':  7.896,
        'Ni':  7.633,
        }

    # degeneracy of states, for saha ionization equation
    SAHA_G0 = {
        'H' :  2.0,
        'He':  1.0,
        'C' :  9.3,
        'N' :  4.0,
        'O' :  8.7,
        'Ne':  1.0,
        'Na':  2.0,
        'Mg':  1.0,
        'Al':  5.9,
        'Si':  9.5,
        'S' :  8.1,
        'K' :  2.1,
        'Ca':  1.2,
        'Cr': 10.5,
        'Fe': 26.9,
        'Ni': 29.5,
        }
    SAHA_G1 = {
        'H' :  1.0,
        'He':  2.0,
        'C' :  6.0,
        'N' :  9.0,
        'O' :  4.0,
        'Ne':  5.0,
        'Na':  1.0,
        'Mg':  2.0,
        'Al':  1.0,
        'Si':  5.7,
        'S' :  4.1,
        'K' :  1.0,
        'Ca':  2.2,
        'Cr':  7.2,
        'Fe': 42.7,
        'Ni': 10.5,
        }
    SAHA_G1G0 = {}
    for _e in SAHA_G0:
        SAHA_G1G0[_e] = SAHA_G1[_e] / SAHA_G0[_e]
    del _e


class _AddonDefaults():
    '''stores defaults for PlasmaCalcs. Use DEFAULTS.ADDONS instead.
    (DEFAULTS.ADDONS is an instance of _AddonDefaults(), instantiated at the bottom of defaults.py)
    '''
    # whether to try to load TFBI theory hookup module.
    # note - adjusting this value after importing PlasmaCalcs will have no effect;
    #    only relevant if adjusted before trying to import PlasmaCalcs.
    # 'attempt' --> try it, but if it fails, don't raise an error.
    # True --> try it, and if it fails, raise an error.
    # False --> don't try it.
    LOAD_TFBI = 'attempt'  # 'attempt', True, or False

    # default max number of ions to consider in TFBI theory, before printing a warning.
    # using more ions (e.g., 4, 5, 6+) gets much slower and less accurate.
    #   theory involves solving ratio of polynomials with degree = 4 * number of fluids.
    #   (numerator degree 2 smaller if using "QN" form of dispersion relation.)
    TFBI_MAX_NUM_IONS = 5

    # default tfbi_EBspeed_grid logmin, logmax, and logstep
    TFBI_EBSPEED_LOGMIN = 3
    TFBI_EBSPEED_LOGMAX = 4   # 4 is sufficient for all simulated TFBI suite points.
    TFBI_EBSPEED_LOGSTEP = 0.01

    # default max Mbytes for tfbi_inputs when solving tfbi_vs_EBspeed.
    # the tfbi_inputs Mbytes are checked before setting EBspeed to the EBspeed grid.
    TFBI_EBSPEED_INPUTS_MBYTES_MAX = 0.1

    # threshold for confirming "yes there is tfbi growth predicted here"
    # 0.0 is the theoretical threshold.
    # a small positive value (e.g. 0.001) helps to avoid tiny errors,
    #   and also ensures the growth is "not so slow that the background conditions might change",
    #   e.g. in the chromosphere the background might change after a few seconds;
    #       10 seconds corresponds to a threshold of 0.1.
    TFBI_GROWTH_THRESH = 0.001

    # default legend_kw for klines in growthplot.
    GROWTHPLOT_LEGEND_KW = dict(
        bbox_to_anchor=(0.2, 0.95),
        loc='upper right',
        fontsize='xx-small',
        handlelength=5,
    )

    # default legend_kw for klines in growthplots (plural).
    GROWTHPLOTS_LEGEND_KW = dict(
        bbox_to_anchor=(0.2, 0.8),
        loc='lower right',
        fontsize='xx-small',
        handlelength=5,
    )


class _EppicDefaults():
    '''stores defaults for PlasmaCalcs. Use DEFAULTS.EPPIC instead.
    (DEFAULTS.EPPIC is an instance of _EppicDefaults(), instantiated at the bottom of defaults.py)
    '''
    # how many digits to zfill eppic snapshot numbers with
    H5_SNAP_ZFILL = 6

    # eppic dx,dy,dz = unsafe_dspace * safety. Probably ~1 or a bit larger.
    # e.g. might use unsafe_dspace = electron ldebye.
    # larger values here are LESS safe.
    DSPACE_SAFETY = 1.1

    # eppic dt = smallest timescale * safety. Probably ~1 or a bit smaller.
    # larger values here are LESS safe.
    DT_SAFETY = 0.9

    # eppic Nx,Ny,Nz = unsafe_Nspace * safety. Probably ~10.
    # e.g. might use unsafe_Nspace = expected scale size of physical features (e.g. 1 wavelength)
    # Nx,Ny,Nz is the actual total number of grid cells in each dimension.
    # Nx=nx*nsubdomains; Ny=ny; Nz=nz.
    NSPACE_SAFETY = 5

    # eppic nt = unsafe_nt * safety. Probably ~20.
    # e.g. might use unsafe_nt = expected timescale of physical features (e.g. 1/growthrate)
    # nt is the total number of timesteps.
    NTIME_SAFETY = 40

    # default safety factor for runtime guess.
    # Consider setting requested time in .slurm file to n_processors * runtime guess * safety.
    # Seems like 1 is usually good enough.
    # (Checked on my tfbi2_eff_RUNS and a bunch of other runs.
    #   For the bigger "production" runs, 1 would have always been (just barely) good enough.
    #   For eff_RUNS, ~25% need >1.5, ~15% need >2, ~4% need >2.5, 0% need >2.7.
    RUNTIME_SAFETY = 1.2

    # eppic nout (before rounding) = min(waveprop_nout, growth_nout)
    # waveprop_nout = (waveprop_time / dt) * NOUT_WAVEPROP_SAFETY
    # growth_nout = (timescale_growth / dt) * NOUT_GROWTH_SAFETY
    # Larger values are LESS safe.
    # E.g. NOUT_WAVEPROP_SAFETY=0.25 implies "at least 4 snapshots per 1 wavelength of wave motion."
    # E.g. NOUT_GROWTH_SAFETY=0.1 implies "at least 10 snapshots per 1 e-folding of growth."
    NOUT_WAVEPROP_SAFETY = 0.25
    NOUT_GROWTH_SAFETY = 0.125

    # eppic nout (rounded appropriately) will be a multiple of NOUT_MULTIPLE * max safe_pow2_subcycle.
    # E.g., nout_multiple=5 causes nout to be a multiple of 10,
    #    assuming safe_pow2_subcycle > 1 for at least 1 specie
    #    (subcycle=2^N with N>0 --> subcycle * 5 will be divisible by 10).
    # Use nout_multiple=1 to ignore this value and just use the max safe_pow2_subcycle requirement.
    NOUT_MULTIPLE = 1

    # safety factor for Rosenberg's criterion for quasineutrality:
    #   quasineutrality is "reasonable" when (nusn / wplasma)^2 << 1.
    #   (use '<= EPPIC.ROSENBERG_SAFETY' instead of '<< 1'.)
    ROSENBERG_SAFETY = 0.5

    # number of "sigma" (thermal widths) to include in eppic vdist outputs.
    # e.g. if vout_nsigma = 3, then vdist will include -3 vthermal to +3 vthermal.
    VDIST_NSIGMA = 4

    # safety factor for subcycling in eppic. Larger is safer.
    # when making eppic input deck, will use subcycle = largest 2^N <= (best possible subcycling / safety)
    SUBCYCLE_SAFETY = 1.5

    # rounding to use for npd choices. None --> round to nearest int
    # e.g. use 10 if you want all npd choices to be rounded to the nearest 10.
    NPD_ROUNDING = 10

    # "target cost" of each subcycled species, when picking number of particles (npd).
    # the idea is to use npd(subcycled) = npd(unsubcycled) * npd_mul_cpu_cost * subcycling
    # e.g. if dist 1 has subcycling = 32, and NPD_MUL_CPU_COST = 0.1,
    #     and dist 0 has subcycling = 1, npd=1000,
    #    then dist 1 should target npd = 3200.
    # (might be lower or higher due to rounding; see NPD_MUL_MAX and NPD_MUL_INCREMENT.)
    NPD_MUL_CPU_COST = 0.2

    # maximum npd_mul to use for subcycled species.
    # e.g. if dist 1 has subcycling = 256, and NPD_MUL_CPU_COST = 0.1,
    #     and NPD_MUL_MAX = 5, then use npd_mul of 5 instead of 25.6.
    # None --> no maximum.
    NPD_MUL_MAX = 10

    # default ndim_space when making eppic input dicks via EppicInstabilityCalculator.
    NDIM_SPACE = 2

    # minimum value for nx when making eppic input decks via EppicInstabilityCalculator.
    #   (can vary total Nx via nsubdomains instead of just nx.
    #    if nx is too small, can see "particle jumps more than entire mesh" error, though.
    #    4 is okay for quick runs, 8 is usually safe, 16 is "always" safe.
    #    should probably be a power of 2.)
    NX_MIN = 16

    # maximum value for nsubdomains when making eppic input decks via EppicInstabilityCalculator.
    # if need more Nx than NX_MIN * NSUBDOMAINS_MAX, will use nx > NX_MIN.
    # [TODO] implement auto-check: ncpu / nsubdomains MUST be a whole number,
    #    so if ncpu is known, need to use nsubdomains max = min(NSUBDOMAINS_MAX, 2^N),
    #    with N == number of factors of 2 inside ncpu. (E.g. ncpu=56 == (2^3)*7 --> N=3)
    NSUBDOMAINS_MAX = 512

    # default nptotcelld for the uncycled distribution (not necessarily dist 0),
    #   when making eppic input decks via EppicInstabilityCalculator.
    NPTOTCELLD0 = 100

    # eppic iwrite = nout * IWRITE_NSNAP.
    #   I.e. writing dump every IWRITE_NSNAP snapshots.
    #   (If making input deck via EppicInstabilityCalculator.)
    IWRITE_NSNAP = 100

    # default values for "other" (e.g., numerical choices) global vars.
    # Nx, Nt, dx, dt handled elsewhere (these are physically meaningful choices, usually.)
    GLOB_DEFAULTS = {
        'nout_avg': 2,  # spatially averaging output arrays by this factor in each dimension.
        'hdf_output_arrays': 2,   # output type. Use 2 for parallel. (non-2 options are very old.)
        'npout': 4000,  # fraction of particles to output; irrelevant(?) when hdf_output_arrays == 2.
        'iwrite_nsnap': 50,  # Equivalent to using iwrite = iwrite_nsnap * nout.
        'iread': 0,  # start from t=0 (iread=0) or dump (iread=1)
        'divj_out_subcycle': -1,  # output divj every Nth timestep. -1 --> don't output divj.
        'fwidth': 3,  # k-space "width" for phi low-pass filter function
        'fsteep': 3,  # k-space "steepness" for phi low-pass filter function
    }

    # default values for "other" distribution vars.
    DIST_DEFAULTS = {
        'init_dist': 1,   # how to initialize the distribution. I've been using 1. (-SE 2025/02/10)
        'part_pad': 4.0,  # max number of particles ever on a single processor == initial amount * part_pad.
        'coll_type': 1,   # collision type for each distribution... I've been using 1. (-SE 2025/02/10)
        'pnvx': 64,  # number of points in the x direction when outputting vdist
        'pnvy': 64,  # number of points in the y direction when outputting vdist
        'pnvz': 32,  # number of points in the z direction when outputting vdist
        # output subcycles: output some info only every Nth snapshot.
        'vdist_out_subcycle': 64,  # velocity distributions. Info I'm not using, hence large number. (-SE 2025/02/10)
        'part_out_subcycle': 1,   # density info - probably want this every snapshot!
        'flux_out_subcycle': 8,   # I don't need flux or nvsqr info nearly as often as density info;
        'nvsqr_out_subcycle': 8,  #  subcycling these also significantly reduces output filesize. (-SE 2025/02/10)
    }


DEFAULTS = _Defaults()
DEFAULTS.PLOT = _PlotDefaults()  # e.g., can access CAX_PAD via DEFAULTS.PLOT.CAX_PAD
DEFAULTS.PHYSICAL = _PhysicalDefaults()  # e.g., can access M_AMU via DEFAULTS.PHYSICAL.M_AMU
DEFAULTS.ADDONS = _AddonDefaults()
DEFAULTS.EPPIC = _EppicDefaults()
